import itertools
import json
import math
from numbers import Number

from table_column import TableColumn
from table import _Table
import formatted_text

class TableBuilder(object):
    """ High level API to build a table in analysis

    :param object data: The data to store in the table. Accepts `pandas.DataFrame`, `numpy.ndarray`,  a list of floats, or a list of strings
    :param str title: table title
    :param int default_rows_per_page: the default number of rows per page when the table is initially rendered
    :param list column_names: names of each column specified as a list of strings
    :param bool striped: whether or not the table rows should be rendered in alternating grey and white stripes
    :param str search_box_place_holder: the default text shown in the search box when user hasn't typed any search keyword
    :param bool page_controls_and_search_visible: if true show a table header with next/previous page and search controls enabled
    """

    ALLOWED_ROWS_PER_PAGE = [10, 20, 50, 100]

    def __init__(self, data, title, default_rows_per_page = 20, column_names = None, striped = True, search_box_place_holder = 'Search', page_controls_and_search_visible=True):
        self._table_id = None

        self.title = title
        self.striped = striped
        self.search_box_place_holder = search_box_place_holder

        if not (default_rows_per_page in self.ALLOWED_ROWS_PER_PAGE):
            raise RuntimeError("Only one of {0} is allowed for default_rows_per_page".format(self.ALLOWED_ROWS_PER_PAGE))
        self.default_rows_per_page = default_rows_per_page

        self._page_controls_and_search_visible = page_controls_and_search_visible

        self._columns = []
        self._col_name_to_idx = {}

        if type(data).__name__ == 'DataFrame':  # treat it as Pandas DataFrame

            if column_names is None:
                column_names = data.columns.values

            # Turning into numpy ndarray
            data = data.values

        if type(data).__name__ == 'ndarray':  # treat it as Numpy Array, need to transpose the data
            if data.ndim != 2:
                raise RuntimeError("Only two dimensional arrays are supported as acceptable values.")

            n_cols = data.shape[1]

            if column_names is None:
                column_names = [str(c) for c in range(n_cols)]

            # Turning into usual Python array.
            final_data = []
            for i in xrange(0, n_cols):
                # Replace Nan, Inf, and -Inf if the value is a number
                # because neither server nor UI can deal with it.
                filtered_col_data = [None if isinstance(e, Number) and (math.isnan(e) or math.isinf(e)) else e for e in data[:,i].tolist()]
                final_data.append(filtered_col_data)
            data = final_data

        n_cols = len(data)
        if column_names is None or len(column_names) != n_cols:
            raise RuntimeError("The number of column names must be the same as number of columns")

        for i in xrange(n_cols):
            self._add_col(data[i], i, column_names[i])


    @property
    def _component_type(self):
        return "TABLE"

    def _add_col(self, col_data, col_idx, col_name):
        self._columns.append(TableColumn(self, list(col_data), col_name))
        self._col_name_to_idx[col_name] = col_idx

    def _change_col_name(self, old_name, new_name):
        col_idx = self._col_name_to_idx[old_name]
        del self._col_name_to_idx[old_name]
        self._col_name_to_idx[new_name] = col_idx

    def __getitem__(self, idx):
        if isinstance(idx, basestring):
            return self._columns[self._col_name_to_idx[idx]]
        else:
            return self._columns[idx]

    def _get_table_data(self, analysis):
        all_cols_data = []  # all columns including the columns for rendering and sorting, each element is a column
        all_cols_name = []  # 1 to 1 mapping to the all_cols_data, each element is the name of a column
        row_components_col = []  # the "row_components" column, each element is the list of components for the a row
        for col_obj in self._columns:
            cols_data, cols_name, cols_comp = col_obj._get_data_columns(analysis)
            assert len(cols_data) == len(cols_name)
            all_cols_data.extend(cols_data)
            all_cols_name.extend(cols_name)

            # for each row, accumulate the new components into the existing row_components column
            if cols_comp:
                if row_components_col:
                    row_components_col = [existing_comp+new_comp for existing_comp, new_comp in itertools.izip(row_components_col, cols_comp)]
                else:
                    row_components_col = cols_comp

        # the "row_components" is just another column in the table data
        all_cols_data.append(row_components_col)
        all_cols_name.append("row_components")
        assert len(all_cols_data) == len(all_cols_name)

        # access the data by row, and turn it into a list of dict, each representing a row
        rows = itertools.izip_longest(*all_cols_data)
        table_data_json = [dict(itertools.izip(all_cols_name, row)) for row in rows]

        return table_data_json

    def _get_table_component(self, table_id, analysis = None):
        title_comp_ref, _ = formatted_text._get_component_ref_and_defs_for_value(
                self.title, analysis, associate_with_analysis=True)  # components in table title is put into the analysis' main component list

        columns = []
        filter_columns = []
        search_columns = []
        for col_obj in self._columns:
            if col_obj.filterable:
                filter_columns.append({
                    "key": col_obj._column_name,
                    "label": col_obj.filter_name })
            if col_obj.searchable:
                search_columns.append(col_obj._column_name)
            if not col_obj.filter_only:
                header_comp_ref, _ = formatted_text._get_component_ref_and_defs_for_value(
                        col_obj.column_header, analysis, associate_with_analysis=True)  # components in table title is put into the analysis' main component list
                columns.append({
                    "displayKey": col_obj._rendered_col_name,
                    "name":  header_comp_ref,
                    "sortKey": col_obj._sort_col_name,
                    "width": col_obj.width })

        # put the default_rows_per_page as the first element, this is what WebUI requires
        page_sizes = [self.default_rows_per_page] + sorted(set(self.ALLOWED_ROWS_PER_PAGE) - {self.default_rows_per_page})
        search_box_place_holder = self.search_box_place_holder
        # if we want to hide the search box and page arrows, send no page sizes to frontend
        if not self._page_controls_and_search_visible:
            page_sizes = None
            search_columns = None
            search_box_place_holder = None

        default_sort_order = {'key': columns[0]["sortKey"], 'order': 'ASC'}

        return _Table(
                table_id = table_id,
                columns = columns,
                defaultSort = default_sort_order,
                filters = filter_columns,
                searchBoxAttributes = search_columns,
                pageSizes = page_sizes,
                searchBoxPlaceholder = search_box_place_holder,
                striped = self.striped,
                title = title_comp_ref)

    def _pre_associate(self, analysis):
        # Required to match the `_Component` interface
        endpoint = '/analysis/%s/tables' % analysis._id
        resp = analysis._eureqa._session.execute(endpoint, method='POST', args=self._get_table_data(analysis))
        table_id = resp['table_id']

        # upload the table component
        table_component = self._get_table_component(table_id, analysis = analysis)
        table_component._pre_associate(analysis)

        self._table_component = table_component
        self._table_id = table_id

    def _associate(self, analysis):
        # Required to match the `_Component` interface
        if not hasattr(self, "_table_component"):
            self._pre_associate(analysis)
        self._table_component._associate(analysis)

    @property
    def _component_id(self):
        return self._table_component._component_id

    def _register(self, analysis):
        # Required to match the `_Component` interface
        # book keeping
        self._analysis = analysis
        self._table_component._register(analysis)

    def _walk(self):
        if hasattr(self, "_table_component"):
            for c in self._table_component._walk():
                yield c
        else:
            yield self

    def _walk_children(self):
        if hasattr(self, "_table_component"):
            for c in self._table_component._walk_children():
                yield c
        else:
            yield self

    def _to_json(self):
        if hasattr(self, "_table_component"):
            return self._table_component._to_json()
        else:
            return {"component_type": "TABLE"}

    def _associate_with_table(self, analysis):
        self._associate(analysis)
        return [self._table_component._to_json()]



from eureqa import Eureqa
from eureqa.analysis_templates.parameters_values import ParametersValues

import sys,imp,os
import argparse
import tempfile
import warnings
import shutil
import importlib
import zipfile
from cStringIO import StringIO
import traceback
import os

class Analysis_template_execution:
    pass

class LocalFatalException(Exception):
    """
    Class that represents fatal exceptions from the local runner
    """
    pass

class Local_analysis_template_execution(Analysis_template_execution):
    """
    Stubs out the analysis_template_execution to provide the same interface while running locally.

    :param Eureqa eureqa: A eureqa connection.
    :param str parameters: JSON string of parameters.
    """
    def __init__(self, eureqa, parameters=None):
        self.template_id = 'local_template_id'
        self.analysis_id = 'local_analysis_id'
        if parameters is None: parameters = {}

        try:
            self.parameters = ParametersValues._from_json(eureqa, parameters, execution=None).parameters
        except:
            warnings.warn("Invalid ParametersValues expression: %s.  Continuing with no ParametersValues." % repr(parameters))

        self._done = False
        self._eureqa = eureqa
        self._analysis = eureqa.create_analysis('Analysis from analysis template', 'This analysis was run from a local template executor.')
        self._first_fatal_exception = None

    def get_analysis(self):
        return self._analysis

    def update_progress(self, message):
        """ Update progress of the running analysis.

        :param str message: Message to print indicating progress.
        """
        print message

    def _report_completion(self):
        self._done = True
        pass

    def _report_validation_completion(self):
        pass

    def _is_waiting_on_confirmation(self):
        return False

    def _is_done(self):
        return self._done

    def report_fatal_error(self, error):
        """ Report a fatal error of the running analysis.

        :param str error: Message to print indicating progress.
        """
        if (not self._first_fatal_exception):
            self._first_fatal_exception = LocalFatalException(error)
        raise self._first_fatal_exception

    def throw_if_fatal_exception(self):
        """
        Throw any error encountered as a fatal exception
        """
        if self._first_fatal_exception:
            raise self._first_fatal_exception

class analysis_template_runner:
    """
    Bootstrapper for analysis templates. Sets up the environment
    required to invoke the analysis template and then invokes it.
    """
    def __init__(self):
        self.tempdir = None
        self.path_addition = None

    def __del__(self):
        if (self.tempdir):
            shutil.rmtree(self.tempdir)

    def _add_to_python_path(self, addition):
        """ add the specified location to the python path """
        # Put the user module directory on the sys.path list, so that
        # any submodules referenced within it can be found.
        #
        # By putting it on the *end* of sys.path, we prevent users
        # from overriding any modules that are already available on
        # the system by including them in their .zip file.
        self.path_addition = addition # ensure we clean it up
        # It is important to insert the module path in front of other modules
        # so it does not load other things like built-in analysis tempaltes.
        sys.path.insert(0, addition)



    def _invoke_analysis_template(self, main_module, eureqa, template_execution):
        """Driver method for analyis template module execution.

        main_module is the main_module (object) for the analysis
        template (has run_analysis methods defined, etc)

        """
        try:
            validate_parameters = getattr(main_module, 'validate_parameters', None)

            if (validate_parameters and not callable(validate_parameters)):
                raise Exception('validate_parameters is not a function in ' + getattr(main_module, '__name__', 'unknown'))

            # run validate_parameters function
            try:
                if (validate_parameters):
                    validate_parameters(eureqa=eureqa, template_execution=template_execution)

            except:
                template_execution.report_fatal_error(traceback.format_exc())
                template_execution._report_completion()
                return


            # if we are need to wait on user confirmation of
            # validation, poll here Note that isn't a busy loop:
            # non-obviously, calling is_waiting_on_confirmation will
            # enforce a 1 sec delay between calls
            template_execution._report_validation_completion()

            while (template_execution._is_waiting_on_confirmation()):
                pass

            # If there is a template execution with errors (exceptions
            # or reported), it will be either errored or aborted
            if (template_execution._is_done()):
                return

            try :
                create_analysis = getattr(main_module, 'create_analysis')
            except Exception:
                raise Exception('Can not find create_analysis in ' + getattr(main_module, '__name__', 'unknown'))


            if (not callable(create_analysis)):
                raise Exception('create_analysis is not a function in ' + getattr(main_module, '__name__', 'unknown'))

            # run create_analysis function
            try:
                create_analysis(eureqa=eureqa, template_execution=template_execution)
            except:
                template_execution.report_fatal_error(traceback.format_exc())


            template_execution._report_completion()

        except:
            try :
                template_execution.report_fatal_error(traceback.format_exc())
            except:
                pass


    def get_analysis_module_from_template(self, eureqa, template):
        """Retrieves the bytes that represent an analysis template from the
        eureqa server (as base64-encoded .zip file module); then
        function, raising an Exception if an error occurs.

        :param Eureqa eureqa: A eureqa connection.
        :param Template template: The Template object containing the desired analysis.
        """

        self.tempdir = tempfile.mkdtemp()
        zip_file = os.path.join(self.tempdir, "analysis_module.zip")
        main_module_name = template.get_module(zip_file)

        # Extract the .zip file to the local filesystem
        # so that modules containing .so files, etc,
        # can run properly
        with zipfile.ZipFile(zip_file,"r") as zip_f:
            zip_f.extractall(self.tempdir)

        # Remove the .zip file; we're done with it
        os.remove(zip_file)

        self._add_to_python_path(self.tempdir);
        import importlib
        return importlib.import_module(main_module_name)

    def run_args(self, arguments, _local_execution_parameters=None):
        """ Run a template with the provided arguments.

        :param str arguments: Command line-style argument string.
        """
        """ Parses a set of command line arguments and performs the appropriate action """

        parser = argparse.ArgumentParser(prog='run_analysis_template',
                                         description="Nutonian Analysis Template Runner. "
                                         "Invokes an analysis template function either "
                                         "from the command line or from Eureqa itself")

        parser.add_argument('-e', '--url', type=str,
                            help='Base url of Eureqa instance',
                            default='http://localhost:10002')
        parser.add_argument('--save-credentials',
                            dest="save_credentials", action="store_true",
                            help='Save password to the Eureqa API credential store (not encrypted)')
        parser.add_argument('--quiet',
                            action="store_true",
                            help="Suppress interactive and debugging output")
        parser.add_argument('-u', '--user_name', type=str,
                            help='Username for connecting to Eureqa',
                            default='jlpicard')
        parser.add_argument('-p', '--password', type=str,
                            help='Password for connecting to Eureqa',
                            default='Changeme1')
        parser.add_argument('-k', '--api_key', type=str,
                            help='API Key for connecting to Eureqa',
                            default=None)
        parser.add_argument('-o', '--organization', type=str,
                            help='Organization within Eureqa',
                            default='nutonian')

        local_args = parser.add_argument_group("Local-Mode Arguments")
        local_args.add_argument('-m', '--module_name',
                                 help="[Local Mode] The path to the python module (file or directory) to use",
                                 dest="module_name")

        subproc_args = parser.add_argument_group("Subprocess-Mode Arguments")
        subproc_args.add_argument('--session_key', type=str,
                                  help='[Subprocess Mode Only] Credentials to use to connect to Eureqa')
        subproc_args.add_argument('--analysis_template_id', type=str,
                                  help='[Subprocess Mode Only] The template id of the analysis template to invoke')
        subproc_args.add_argument('--execution_id', type=str,
                                  help='[Subprocess Mode Only] The execution id of the analysis template which was invoked')


        args = parser.parse_args(arguments)

        eureqa = Eureqa(url=args.url, user_name=args.user_name, password=args.password, key=args.api_key, save_credentials=args.save_credentials,
                        organization=args.organization, verify_ssl_certificate=False, session_key=args.session_key, interactive_mode = not args.quiet)

        # Get the appropriate entrypoint
        local_template_execution = None
        if (args.analysis_template_id):
            template = eureqa._get_analysis_template(args.analysis_template_id)
            template_execution = template._get_execution(args.execution_id)
            analysis_module = self.get_analysis_module_from_template(eureqa, template)
        elif (args.module_name):
            local_template_execution = Local_analysis_template_execution(eureqa, _local_execution_parameters)
            template_execution = local_template_execution
            self._add_to_python_path(args.module_name)
            analysis_module = __import__(args.module_name)
        else:
            raise Exception('Must supply either --module_name or --analysis_template_id')

        # now, run the the method
        self._invoke_analysis_template(analysis_module, eureqa, template_execution)

        if (local_template_execution):
            local_template_execution.throw_if_fatal_exception()


# launch using program arguments if main file
if (__name__ == "__main__"):
    runner = analysis_template_runner()
    runner.run_args(sys.argv[1:])



import uuid
import itertools
import math

import formatted_text
from base import _Component

class TableColumn(object):
    """ Represent a column in table

    :param ~eureqa.analysis.components.table_builder.TableBuilder parent_table: the containing table
    :param list col_data: data for this column. Can be either a list of str or a list of float
    :param str col_name: name of this column
    """

    def __init__(self, parent_table, col_data, col_name):
        self._parent_table = parent_table
        self._col_data = col_data
        self._column_name = col_name
        self.column_header = col_name
        self._sort_values = None
        self._sort_col_name = col_name
        self._rendered_values = None
        self._rendered_col_name = col_name
        self.searchable = True
        self._width = 1
        self.filterable = False
        self._filter_only = False
        self.filter_name = col_name

    @property
    def column_name(self):
        """The name of this column"""

        return self._column_name

    @column_name.setter
    def column_name(self, new_name):
        """Change the name of this column

        :param str new_name: The new name of the column
        """

        try:
            self._parent_table._change_col_name(self._column_name, new_name)
        except KeyError:
            pass  # if this is an orphan column, ignore and continue

        if self._rendered_col_name == self._column_name:
            self._rendered_col_name = new_name
        if self._sort_col_name == self._column_name:
            self._sort_col_name = new_name
        self._column_name = new_name

    @property
    def sort_values(self):
        """A list of values specifying how the column is sorted"""

        return self._sort_values

    @sort_values.setter
    def sort_values(self, sort_values):
        """Change how the column is sorted

        Note the input must be a list of numbers. The input could be a pandas.Series only if the index runs from 0 to (n-1)

        :param a list of numbers sort_values: the value for each row used for sorting
        """

        self._sort_col_name = str(uuid.uuid4())

        # Backend expects no nans/infs, so we replace nan/-inf with (min(values)-1) and +inf with (max(values)+1).
        # Also we cap the magnitude of all values at a value which is smaller than the largest value JS can represent.
        self._sort_values = sort_values
        largest_value = 2**52
        smallest_value = -largest_value
        set_to_min = []
        set_to_max = []
        for i in xrange(len(self._sort_values)):
            value = self._sort_values[i]
            # ignore strings -- the processing we need to do only applies to numeric values
            if not isinstance(value, float) and not isinstance(value, int):
                continue
            if (math.isnan(value) or (math.isinf(value) and value < 0)) or (value < smallest_value):
                set_to_min.append(i)
            elif (math.isinf(value) and value > 0) or (value > largest_value):
                set_to_max.append(i)
        for i in set_to_min:
            self._sort_values[i] = smallest_value
        for i in set_to_max:
            self._sort_values[i] = largest_value

    @property
    def rendered_values(self):
        """A list of values specifying how the column is rendered"""

        return self._rendered_values

    @rendered_values.setter
    def rendered_values(self, rendered_values):
        """Change how the column is rendered

        :param a list of str or numbers: the value for each row used for rendering
        """
        self._rendered_values = rendered_values
        self._rendered_col_name = str(uuid.uuid4())

    @property
    def width(self):
        """ The width of this column"""

        return self._width

    @width.setter
    def width(self, width):
        """ Change the width of this column

        :param double width: a number representing the percentage of whole table width for this column
        """

        if width <= 0:
            raise RuntimeError("Column width must be larger than zero")
        self._width = width

    @property
    def filter_only(self):
        """ Whether this column is only for filtering, if True this column doesn't appear in the table"""

        return self._filter_only

    @filter_only.setter
    def filter_only(self, filter_only):
        """ Change whether this column is only for filtering

        :param bool filter_only: whether or not this column should be set to filter only
        """

        self._filter_only = filter_only
        self.filterable = True

        # filter_only columns don't show up in the table component, it would be
        # very confusing if you can search on them
        self.searchable = False

    def _get_data_columns(self, analysis):
        cols_name = [self._column_name]
        cols_data = [self._col_data]
        cols_comp = []

        if self._sort_values is not None:
            cols_name.append(self._sort_col_name)
            cols_data.append(self._sort_values)

        if self._rendered_values is not None:
            comp_ref_and_comp_def_by_row = [
                    formatted_text._get_component_ref_and_defs_for_value(rendered_value, analysis, associate_with_analysis=False)
                    for rendered_value in self._rendered_values]
            comp_ref, comp_def = itertools.izip(*comp_ref_and_comp_def_by_row)  # transpose it into 2 columns

            cols_name.append(self._rendered_col_name)
            cols_data.append(comp_ref)
            cols_comp = comp_def

        return cols_data, cols_name, cols_comp



#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from parameter import Parameter
from text_parameter import TextParameter
from data_source_parameter import DataSourceParameter
from top_level_model_parameter import TopLevelModelParameter
from variable_parameter import VariableParameter
from data_file_parameter import DataFileParameter
from combo_box_parameter import ComboBoxParameter
from numeric_parameter import NumericParameter

import json
from collections import OrderedDict

class Parameters:
    """Analysis template parameters definition

    :param list[Parameter] parameters: The list of parameters for the template, whether text, variable or datasource.

    """

    def __init__(self, parameters=None):
        """Initializes a new instance of the ~Parameters class
        """
        self.parameters = OrderedDict((p.id, p) for p in parameters) if parameters is not None else parameters

    def _to_json(self):
        body = {}
        if self.parameters is not None:
            body['parameters'] = [x._to_json() for x in self.parameters.itervalues()]
        return body

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    _parameter_map = {
        'combo_box': ComboBoxParameter,
        'data_file': DataFileParameter,
        'datasource': DataSourceParameter,
        'numeric': NumericParameter,
        'top_level_model': TopLevelModelParameter,
        'text': TextParameter,
        'text_multiline': TextParameter,
        'variable': VariableParameter
    }

    @classmethod
    def _from_json(cls, body):
        ret = Parameters()
        if not 'parameters' in body: return ret

        parameters = []
        for element in body['parameters']:
            parameter = Parameter._from_json(element)
            if not parameter._type in cls._parameter_map: raise Exception('Invalid analysis template parameter type \'%s\' for parameter \'%s\' with label \'%s\'' %(parameter._type, parameter.id, parameter.label))
            parameter = cls._parameter_map[parameter._type]._from_json(element)
            parameters.append(parameter)
        ret.parameters = OrderedDict((p.id, p) for p in parameters)
        return ret



#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from parameter import Parameter

import json

class DataFileParameter(Parameter):
    """Data file parameter description for analysis template

    :param str id: The id of the parameter that will be passed together with its value to an analysis script.
    :param str label: The parameter label that will be shown in UI.
    :param str description: A description that will be shown in UI.
    :param list[str] filetypes: The list of accepted filetypes.

    :var str eureqa.analysis_templates.DataFileParameter.id: The id of the parameter that will be passed together with its value to an analysis script.
    :var str eureqa.analysis_templates.DataFileParameter.label: The parameter label that will be shown in UI.
    :var str eureqa.analysis_templates.DataFileParameter.description: A description that will be shown in UI.
    :var list[str] filetypes: The list of accepted filetypes.
    """

    def __init__(self, id, label, description, filetypes):
        """Initializes a new instance of the ~DataFileParameter class
        """
        Parameter.__init__(self, id, label, "data_file")
        self.description = description
        self.filetypes = filetypes

    def _to_json(self):
        body = Parameter._to_json(self)
        if self.description is not None:
            body["description"] = self.description
        if self.filetypes is not None and len(self.filetypes) > 0:
            body['filetypes'] = [x for x in self.filetypes]
        return body

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    @staticmethod
    def _from_json(body):
        ret = DataFileParameter(None, None, None, None)
        Parameter.from_json(ret, body)
        if ret._type != "data_file": raise Exception("Invalid type '%s' specified for data_file parameter" % ret._type)
        if 'description' in body:
            ret.description = body["description"]
        if 'filetypes' in body:
            ret.filetypes = [x for x in body['filetypes']]
        return ret



import sys

from collections import defaultdict

from .install_analysis_template import main as install_main
from .run_analysis_template import main as run_main

def error(args):
    print "Unrecognized command"

if len(sys.argv) < 2:
    error(sys.argv)

command = sys.argv[1]
remaining_args = [sys.argv[0]] + sys.argv[2:]
    
actions = defaultdict(lambda: error,
                      { 'run_analysis_template': run_main,
                        'install_analysis_template': install_main })

actions[command](remaining_args)



eureqa_colors = {
    "WHITE": "#fff",
    "GREEN": "#62B331",
    "ORANGE": "#ec7e27",
    "YELLOW": "#FFD61C",
    "RED": "#bb1c1a",
    "GOLD": "#edb619",
    "PURPLE": "#8D5E9A",
    "PRIMARY_BLUE": "#238CBC",
    "LINK_BLUE": "#069",
    "LIGHT_GREY": "#ddd",
    "GREY": "#A9A5A0",
    "MEDIUM_GREY": "#777",
    "DARK_GREY": "#494745",
    "ICON_GREY": "#494745",
    "DEFAULT_FONT": "#333333",
    "LIGHT_FONT": "#777",
    "BACKGROUND": "#f4f4f4",
    "PRIMARY_BUTTON": "#ec7e27",
    "SECONDARY_BUTTON": "#A9A5A0",
    "BOX_SHADOW": "0px 2px 3px 0px #ddd",
    "BORDER_LIGHT": "1px solid #ddd",
    "BORDER_MEDIUM": "1px solid #777",
    "BORDER_DOTTED": "1px dotted #777",
    "BORDER_HIGHLIGHT": "1px solid #238CBC",
    "TABLE_ROW_SELECTED": "1px double #238CBC"
}


#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from eureqa.analysis_cards.analysis_card import AnalysisCard

import json
import types

class Button(object):
    """ **BETA**
    An HTML button inside an HtmlCard

    :param str title: Title of the card
    """

    class Events:
        """ Constants that identify the types of events that can trigger an action """
        CLICK = "click"
        DBLCLICK = "dblclick"
        MOUSEUP = "mouseup"
        MOUSEDOWN = "mousedown"
        MOUSEENTER = "mouseenter"
        MOUSELEAVE = "mouseleave"

    def __init__(self, title, _actions=None):
        """ init """
        self.title = title
        self._actions = _actions if _actions else []

    def add_action(self, card_action, args=None, kwargs=None, event=Events.CLICK):
        """ Add an action to this button, to be performed when the specified event happens

        :param instancemethod card_action: Method on an analysis-card instance to invoke.  For example, `card.replace`.
        :param tuple args: Arguments to the `card_action` method
        :param dict kwargs: Keyword arguments to the `card_action` method
        :param str event: Event that triggers this action
        """
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        assert hasattr(card_action, 'im_class') and issubclass(card_action.im_class, AnalysisCard),\
            "'card_action' must be a member function on an analysis-card class"
        assert hasattr(card_action, 'im_func') and hasattr(card_action, 'im_self'),\
            "'card_action' must be a member function on an analysis-card class"
        assert len(args) == card_action.im_func.func_code.co_argcount - 1,\
            "Expected %d arguments; got %d" % (card_action.im_func.func_code.co_argcount - 1, len(args))

        action_map = {
            card_action.im_class.replace.im_func: "replaceCardData"
        }

        if card_action.im_func not in action_map:
            raise AssertionError,\
                "'card_action' argument '%s.%s' is not a supported callback function.  Must be one of: (%s)"  \
                   % (card_action.im_class.__name__, card_action.__name__,
                      ', '.join('%s.%s' % (card_action.im_class.__name__, x.__name__) for x in action_map.iterkeys()))

        self._actions.append(_Action(args[0], event, card_action.im_self._private_channel, action_map[card_action.im_func]))

    def add_replace_action(self, target_card, replacement_card, event=Events.CLICK):
        """ Add a 'Replace' action to the specified card.
        Equivalent to `add_action(target_card.replace, (replacement_card,), event=event)`.

        :param eureqa.analysis_card.AnalysisCard target_card: Card to be replaced
        :param eureqa.analysis_card.AnalysisCard replacement_card: Card to replace `target_card` with
        :param Events event: Event to trigger the replacement
        """
        self._actions.append(_Action(replacement_card, event, target_card._private_channel, "replaceCardData"))

    def to_html(self, html_tag='a', **kwargs):
        """ Render this button as an HTML tag

        For example:  to_html('Go!', html_tag='button', style='color: blue;') would return
        "<button style='color: blue;' (...)>Go!</button>"

        :param str html_tag: Tag name.  For example, 'a', 'input', 'button', 'div'
        :param \**kwargs: Tag parameters.  For example, "style='color: blue;'"
        :return: HTML string representation of this button
        """

        # Hack/workaround for HTML links
        # Links don't respond helpfully to clicks unless they have a 'href'
        if html_tag == 'a' and 'href' not in kwargs:
            kwargs['href'] = '#'

        # '&#39;' is the HTML escape code for single-quote
        # HTML allows property strings to be wrapped in either single or double quotes.
        # Use single quotes:  At least one of these properties is a big JSON blob;
        # escaped JSON is much more readable if double-quote is not a special character.
        params = ' '.join("%s='%s'" % (key, value.replace("'", '&#39;')) for key, value in kwargs.iteritems())

        data = json.dumps(self.to_json()['actions']).replace("'", '&#39;')

        return "<%(html_tag)s%(param_space)s%(params)s data-analysis_events='%(data)s'>%(title)s</%(html_tag)s>" % {
            'html_tag': html_tag,
            'param_space': ' ' if params else '',
            'params': params,
            'data': data,
            'title': self.title
        }

    def to_json(self):
        return {
            'title': self.title,
            'actions': [x.to_json() for x in self._actions]
        }

    @classmethod
    def from_json(cls, data):
        return cls(data['title'], [_Action.from_json(x) for x in data['actions']])


class _Action(object):
    def __init__(self, payload, event, channel, action):
        self.payload = payload
        self.event = event
        self.channel = channel
        self.action = action

    def to_json(self):
        """
        :return: This _Action, represented as primitive JSON-serializable objects
        """
        return {
            'payload': self.payload.to_json(),
            'event': self.event,
            'channels': [ str(self.channel) ],
            'action': self.action
        }

    @classmethod
    def from_json(cls, data):
        assert len(data['channels']) in [0, 1]
        return cls(AnalysisCard.from_json(data['payload']), data['event'], data['channels'][0], data['action'])




import sys, os

## Add the Eureqa API and the test runner to sys.path
## Modify these lines as needed.
sys.path.append(os.path.abspath("../.."))
sys.path.append(os.path.abspath(".."))

from eureqa import *
from eureqa.analysis_templates import *
from eureqa.utils.modulefiles import open_module_file

## This function is required for the analysis template.
## Modify it to create any analyses that you need.
def create_analysis(eureqa, template_execution):
    template_execution.update_progress('Starting running analysis template.')
    analysis = template_execution.get_analysis()
    analysis.name = 'Test analysis template execution 1'

    ## Open the "sample.csv" file contained within this module.
    ## When running on the remote server, this fetches the file.
    with open_module_file('sample.csv') as f:
        data_source = eureqa.create_data_source('Analysis_card__data_source', f)
        
    input_variables = set(data_source.get_variables())-{'Weight'}
    settings = eureqa.search_templates.numeric('Test analysis templates 1', 'Weight', input_variables)
    template_execution.update_progress('Creating search.')
    search = data_source.create_search(settings)
    template_execution.update_progress('Starting search')
    search.submit(1000)
    while len(search.get_solutions()) < 2 and search.is_running:
        pass
    template_execution.update_progress('Found 2 models. Stopping search.')
    search.stop()
    template_execution.update_progress('Creating cards.')

    card_title = template_execution.parameters['example_text_param'].value
    for solution in search.get_solutions():
        analysis.create_model_card(solution, card_title)

    template_execution.update_progress('Done!')


## This function is optional.
## If provided, the template will be a parameterized template;
## in order to run your template, Eureqa users will be prompted to provide these values.
def template_parameters():
    return Parameters(
            [TextParameter('example_text_param','Input Text'),
             TopLevelModelParameter('example_model_param','Select a Model', None),
             DataSourceParameter('example_datasource_param', 'Select a DataSource',
                                 {VariableParameter('variable_param_1', 'Variable')})])

## This section is not required for an analysis template,
## but can be a convenience for developers.
## It causes the module to be runnable; and when run, to
## behave as if you had invoked run_analysis_template.py to
## run the template.
## Note -- to run this, the directory containing
## 'run_analysis_template.py' must be in PYTHONPATH.
if __name__ == "__main__":
    from eureqa.analysis_templates.runner.client import main

    ## By default, take any arguments from the command line.
    args = list(sys.argv)
    ## Alternatively, if you want to construct the command line from scratch:
    #args = ["run_analysis_template.py"]

    ## Construct a Parameters object representing the template's parameters.
    ## These can be difficult to construct at the command line;
    ## constructing them in Python can be easier.
    import json
    parameter_values = ParametersValues(
            [TextParameterValue(None, 'example_text_param','Text entered by the user'),
             TopLevelModelParameterValue(None, 'example_model_param', "a + 2", None, None, None),
             DataSourceParameterValue(None, 'example_datasource_param', 'value of datasource',
                {VariableParameterValue(None, 'variable_param_1', 'value of variable')})])
    args += ["--parameter-values", json.dumps(parameter_values._to_json())]

    ## The one required argument to run_analysis_template.py
    ## is the path to ourselves.
    ## Python knows that, so add it automatically.
    args += [os.path.abspath(os.path.dirname(__file__))]
    
    main(args)



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from eureqa.analysis.components.base import _Component
from eureqa.utils.data_holder import DataHolder
from eureqa.data_source import DataSource

import warnings

class CustomPlot(_Component):
    """
    Represents a plot to be displayed in an analysis plot card.

    By calling the plot() method multiple times on one instance of this class, multiple lines and scatter plots can
    be superimposed on this object.

    :param DataSource datasource: The data source containing the data to be plotted.
    :param str width: Set manual dimensions for the plot in "px" units (e.g. 350px). Defaults to full panel width.
    :param str height: Set manual dimensions for the plot in "px" units. Defaults to a constant height of 400px.
    :param str x_axis_label: The label to use on the x axis of the plot.
    :param str y_axis_label: The label to use on the y axis of the plot.
    :param bool show_legend: Controls whether or not a legend will be used in the plot.
    :param bool zero_x_line: Draws a horizontal line through the origin
    :param bool zero_y_line: Draws a vertical line through the origin
    :param str x_tick_format: Format for x axis tick labels. Valid values are "date" or anything supported by D3: https://github.com/mbostock/d3/wiki/Formatting#d3_format. Defaults to our internal number format
    :param str y_tick_format: Format for y axis tick labels. Valid values are "date" or anything supported by D3: https://github.com/mbostock/d3/wiki/Formatting#d3_format. Defaults to our internal number format
    :param str guides_type: The type of value-guides to show when hovering over a point in the graph. Valid values are "XY", "YY" or False. XY guides will show the x and y values for the point under the cursor. YY guides will show the x and y values of each component for the data point closest to the cursor. False will turn off value guides.
    """

    _component_type_str = 'CUSTOM_GRAPH'

    def _fields(self):
        return super(CustomPlot, self)._fields() + ['options']

    def __init__(self,
                 datasource=None,
                 width=None,
                 height='400px',
                 x_axis_label=None,
                 y_axis_label=None,
                 show_legend=True,
                 zero_x_line=False,
                 zero_y_line=False,
                 x_tick_format=None,
                 y_tick_format=None,
                 guides_type="XY",
                 _analysis=None,
                 _component_id=None,
                 _component_type=None):
        """ Plot init """
        self._options = {
            'datasource_id': None,
            'plot':{
                'dimensions': {
                    'width': width,
                    'height': height
                },
                'x_axis_label': x_axis_label,
                'y_axis_label': y_axis_label,
                'legend': show_legend,
                'zero_x_line': zero_x_line,
                'zero_y_line': zero_y_line,
                'x_tick_format': x_tick_format,
                'y_tick_format': y_tick_format,
                'num_x_ticks': None,
                'num_y_ticks': None,
                'number_of_rows': 200,
                'guides': {
                    'type': guides_type
                }
        }, 'components':[]}
        if datasource is not None:
            datasource = datasource
        if guides_type is not None: self._options['plot']['guides'] = {'type':guides_type}
        self._data_holder = DataHolder()
        self._var_counter = 0

        super(CustomPlot, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)


    # This property is an ugly hack.
    # Its purpose is to enable the legacy syntax "myCustomPlot.plot.plot()"
    # (for compatibility with older Analysis Cards)
    # while supporting the new flat syntax "myCustomPlot.plot()".
    # If/when we're able to delete it, we should rename the "_plot()" method below to "plot()".
    @property
    def plot(self):
        """
        Add new data to the plot with the specified options.

        If a datasource is specified, x and y can be provided as expressions in terms of the variables in that datasource. Otherwise, x and y must be lists of data of identical length.

        :param list x: X axis data. Either a list of input values, or if a datasource is specified, an expression in terms of that datasource's variables.
        :param list y: Y axis data. Either a list of input values, or if a datasource is specified, an expression in terms of that datasource's variables.
        :param DataSource datasource: If provided, a specific DataSource to use as a base for the variables to be plotted.
        :param str style: A matplotlib-like style string of 'o' or '-' or '-o' to indicate circle, line, or line-circle, respectively.
        :param str color: CSS color.
        :param int line_width: The width of the line, if applicable.
        :param int circle_radius: The radius of each circle, if applicable.
        :param bool use_in_plot_range: The chart auto-computes the x and y axis ranges based on the data for each component. If you want to add a component, but have its data not be used to compute the x and y axis ranges, set this value to False. For example, if you want to make a scatter plot, with a trend line going through it, then setting this field to False for the trend-line component will make the chart snugly fit the points, with the trend line extending beyond.
        :param list error_bars_upper_values: Specifies the tops of error bars. Either a list of values, or if a datasource is specified, an expression in terms of that datasource's variables. If input values is a list, then this must be a list as well.
        :param list error_bars_lower_values: Specifies the bottoms of error bars. Either a list of values, or if a datasource is specified, an expression in terms of that datasource's variables. If input values is a list, then this must be a list as well.
        :param str legend_label: A string label to be used in the legend. If unspecified, this plot will not appear in the legend.
        :param str tooltip_template: Any valid Handlebars template. The template will be provided the x and y values as "x_value" and "y_value", so you can use these to format a tooltip for each point. For example, "<div class="my_awesome_tooltip">When age is {{x_value}}, salary is expected to be {{y_value}}</div>"
        """
        class WrapperClass(object):
            def __init__(self, parent):
                self.__dict__["_parent"] = parent
            def __call__(self, *args, **kwargs):
                return self._parent._plot(*args, **kwargs)
            def __getattr__(self, name):
                warnings.warn("'CustomPlot.plot' field is deprecated.  Use 'CustomPlot' directly instead.",
                              DeprecationWarning)
                return getattr(self.__dict__["_parent"], name)
            def __setattr__(self, name, val):
                return setattr(self._parent, name, val)
            def __delattr__(self, name):
                return delattr(self._parent, name)
        return WrapperClass(self)

    @property
    def width(self):
        """
        The width of the plot.  Represented as a string containing a valid CSS "width" attribute; for example, '400px'.

        :rtype: str
        """
        return self._options.get("dimensions").get("width")

    @width.setter
    def width(self, val):
        self._options["dimensions"]["width"] = val


    @property
    def height(self):
        """
        The height of the plot.  Represented as a string containing a valid CSS "width" attribute; for example, '300px'.

        :rtype: str
        """
        return self._options.get("dimensions").get("height")

    @height.setter
    def height(self, val):
        self._options["dimensions"]["height"] = val
        self._update()


    @property
    def x_axis_label(self):
        """
        The label of the plot's X axis

        :rtype: str
        """
        return self._options.get("x_axis_label")

    @x_axis_label.setter
    def x_axis_label(self, val):
        self._options["x_axis_label"] = val
        self._update()


    @property
    def y_axis_label(self):
        """
        The label of the plot's Y axis

        :rtype: str
        """
        return self._options.get("y_axis_label")

    @y_axis_label.setter
    def y_axis_label(self, val):
        self._options["y_axis_label"] = val
        self._update()


    @property
    def show_legend(self):
        """
        Whether or not the legend should be shown

        :rtype: bool
        """
        return self._options.get("legend")

    @show_legend.setter
    def show_legend(self, val):
        self._options["legend"] = val
        self._update()


    @property
    def zero_x_line(self):
        """
        Whether an axis line should be shown for x=0

        :rtype: bool
        """
        return self._options.get("zero_x_line")

    @zero_x_line.setter
    def zero_x_line(self, val):
        self._options["zero_x_line"] = val
        self._update()


    @property
    def zero_y_line(self):
        """
        Whether an axis line should be shown for y=0

        :rtype: bool
        """
        return self._options.get("zero_y_line")

    @zero_y_line.setter
    def zero_y_line(self, val):
        self._options["zero_y_line"] = val
        self._update()


    @property
    def x_tick_format(self):
        """
        Format for x axis tick labels.
        Valid values are "date" or anything supported by D3:
        https://github.com/mbostock/d3/wiki/Formatting#d3_format. Defaults to our internal number format

        :return: str
        """
        return self._options.get("x_tick_format")

    @x_tick_format.setter
    def x_tick_format(self, val):
        self._options["x_tick_format"] = val
        self._update()


    @property
    def y_tick_format(self):
        """
        Format for y axis tick labels.
        Valid values are "date" or anything supported by D3:
        https://github.com/mbostock/d3/wiki/Formatting#d3_format. Defaults to our internal number format

        :return: str
        """
        return self._options.get("y_tick_format")

    @y_tick_format.setter
    def y_tick_format(self, val):
        self._options["y_tick_format"] = val
        self._update()


    @property
    def _num_x_ticks(self):
        """
        Number of x axis tick labels.
        Defaults to as many as fit comfortably without overlap

        :rtype: int
        """
        return self._options.get("num_x_ticks")

    @_num_x_ticks.setter
    def _num_x_ticks(self, val):
        self._options["zero_y_line"] = val
        self._update()


    @property
    def _num_y_ticks(self):
        """
        Number of y axis tick labels.
        Defaults to as many as fit comfortably without overlap

        :rtype: int
        """
        return self._options.get("num_y_ticks")

    @_num_y_ticks.setter
    def _num_y_ticks(self, val):
        self._options["num_y_ticks"] = val
        self._update()


    @property
    def _max_displayed_points(self):
        """
        Maximum number of data points to sample for each component

        :rtype: int
        """
        return self._options.get("number_of_rows")

    @_max_displayed_points.setter
    def _max_displayed_points(self, val):
        self._options["number_of_rows"] = val
        self._update()


    @property
    def guides_type(self):
        """
        The type of value-guides to show when hovering over a point in the graph.
        Valid values are "XY", "YY" or False.
        XY guides will show the x and y values for the point under the cursor.
        YY guides will show the x and y values of each component for the data point closest to the cursor.
        False will turn off value guides.

        :rtype: str
        """
        return self._options.get("guides").get("type")

    @guides_type.setter
    def guides_type(self, val):
        self._options["guides"]["type"] = val
        self._update()


    @property
    def datasource(self):
        """The data source providing data for this component

        :return: data source providing data for this component
        """
        if self._options.get("datasource_id") and hasattr(self, "_eureqa"):
            return self._eureqa.get_data_source_by_id(self._options["datasource_id"])

    @datasource.setter
    def datasource(self, val):
        self._datasource = val
        self._options["datasource_id"] = val._data_source_id
        self._eureqa = val._eureqa
        self._update()


    def upload_data(self, eureqa=None):
        """
        Upload the plot data to eureqa. This is required before the plot can be viewed.

        :param Eureqa eureqa: a eureqa connection
        """
        if eureqa is None:
            eureqa = self._analysis._eureqa

        csv_file = self._data_holder.get_csv_file()
        datasource = eureqa.create_data_source("Data Holder for Plot", csv_file, _hidden=True)
        self.datasource = datasource
        self._options['datasource_id'] = datasource._data_source_id

    @staticmethod
    def _plot_type(style):
        """
        For internal use only.

        Translate a matplotlib-like style string into one of the supported plot types

        :param str style: a matplotlib-like style string
        :return: internal plot type
        :rtype: str
        """
        if '-' in style and 'o' in style: return 'line_circle'
        if 'o' in style: return 'circle'
        if '-' in style: return 'line'
        return None

    def _plot_data(self, x_data, y_data, config):
        """
        For internal use only.

        Given two lists of data (and perhaps variable names to use for those lists), add the data to the plot with the specified options.

        :param list x_list: a list of input values for the x axis
        :param list y_list: a list of input values for the y axis
        :param str x_name: an optional name for the x variable. If unspecified, default is (Series %i X), where i is the plot number
        :param str y_name: an optional name for the y variable. If unspecified, default is (Series %i Y), where i is the plot number
        """
        if (config.get('error_bars_upper_values') is not None and not isinstance(config.get('error_bars_upper_values'), list))\
        or (config.get('error_bars_upper_values') is not None and not isinstance(config.get('error_bars_upper_values'), list)):
            raise Exception("CustomPlot.plot error: since x and y were inputted as lists, error bars must be provided as lists")

        self._options['datasource_id'] = None

        self._var_counter += 1
        x_name = config.get('x_name', '(Series %i %s)' % (self._var_counter, 'X'))
        y_name = config.get('y_name', '(Series %i %s)' % (self._var_counter, 'Y'))
        self._data_holder.add_column(x_name, x_data)
        self._data_holder.add_column(y_name, y_data)

        error_bars_upper_values_name = config.get('error_bars_upper_values_name')
        error_bars_lower_values_name = config.get('error_bars_lower_values_name')
        if config.get('error_bars_upper_values') is not None and config.get('error_bars_lower_values') is not None:
            error_bars_upper_values_name = error_bars_upper_values_name or '(Series %i upper error bars)' % (self._var_counter)
            error_bars_lower_values_name = error_bars_lower_values_name or '(Series %i lower error bars)' % (self._var_counter)
            self._data_holder.add_column(error_bars_upper_values_name, config.get('error_bars_upper_values'))
            self._data_holder.add_column(error_bars_lower_values_name, config.get('error_bars_lower_values'))

        plot_options = {
            'type': CustomPlot._plot_type(config.get('style')),
            'color': config.get('color'),
            'width': config.get('line_width'),
            'radius': config.get('circle_radius'),
            'x_values': x_name,
            'y_values': y_name,
            'use_in_plot_range': config.get('use_in_plot_range'),
            'error_bars_upper_values': error_bars_upper_values_name,
            'error_bars_lower_values': error_bars_lower_values_name,
            'inside_legend': config.get('legend_label') is not None,
            'legend_label': config.get('legend_label')
        }
        if config.get('tooltip_template') is not None:
            plot_options['tooltip'] = { 'tooltip_template': config.get('tooltip_template') }

        self._options['components'].append(plot_options)

    def _plot_expression(self, x_expr, y_expr, datasource, config):
        """
        For internal use only.

        Given a datasource and two expressions in terms of variables in that datasource, evaluate the expressions and add the data to the plot with the specified options.

        :param list x_expr: an expression for the x axis in terms of the datasource
        :param list y_expr: an expression for the y axis in terms of the datasource
        :param Datasource datasource: a datasource to evaluate the expressions on
        :param config: configuration as dictionary
        """
        if not datasource or not isinstance(datasource, DataSource): raise Exception("CustomPlot.plot error: an existing DataSource object must be provided if x and y are provided as strings referencing variables or expressions")
        if (config.get('error_bars_upper_values') is not None and not isinstance(config.get('error_bars_upper_values'), str))\
        or (config.get('error_bars_upper_values') is not None and not isinstance(config.get('error_bars_upper_values'), str)):
            raise Exception("CustomPlot.plot error: since x and y were inputted as expression strings, error bars must be provided as expression strings")

        x_data, y_data = datasource._eureqa._expression_values(datasource, x_expr, y_expr)
        config['x_name'] = '(Datasource %s: %s)' %(datasource.name, x_expr)
        config['y_name'] = '(Datasource %s: %s)' %(datasource.name, y_expr)
        config['error_bars_upper_values_name'] = None
        config['error_bars_lower_values_name'] = None
        if config.get('error_bars_upper_values') is not None and config.get('error_bars_lower_values') is not None:
            config['error_bars_upper_values_name'] = '(Datasource %s: %s)' %(datasource.name, config.get('error_bars_upper_values'))
            config['error_bars_lower_values_name'] = '(Datasource %s: %s)' %(datasource.name, config.get('error_bars_lower_values'))
            config['error_bars_upper_values'] = datasource._eureqa.evaluate_expression(datasource, config.get('error_bars_upper_values'))[config.get('error_bars_upper_values')]
            config['error_bars_lower_values'] = datasource._eureqa.evaluate_expression(datasource, config.get('error_bars_lower_values'))[config.get('error_bars_lower_values')]
        self._plot_data(x_data, y_data, config)

    def _plot(self, x, y,
             datasource=None,
             style='-',
             color='black',
             line_width=1,
             circle_radius=1,
             use_in_plot_range=True,
             error_bars_upper_values=None,
             error_bars_lower_values=None,
             legend_label=None,
             tooltip_template=None):
        """
        Add new data to the plot with the specified options.

        If a datasource is specified, x and y can be provided as expressions in terms of the variables in that datasource. Otherwise, x and y must be lists of data of identical length.

        :param list x: X axis data. Either a list of input values, or if a datasource is specified, an expression in terms of that datasource's variables.
        :param list y: Y axis data. Either a list of input values, or if a datasource is specified, an expression in terms of that datasource's variables.
        :param DataSource datasource: If provided, a specific DataSource to use as a base for the variables to be plotted.
        :param str style: A matplotlib-like style string of 'o' or '-' or '-o' to indicate circle, line, or line-circle, respectively.
        :param str color: CSS color.
        :param int line_width: The width of the line, if applicable.
        :param int circle_radius: The radius of each circle, if applicable.
        :param bool use_in_plot_range: The chart auto-computes the x and y axis ranges based on the data for each component. If you want to add a component, but have its data not be used to compute the x and y axis ranges, set this value to False. For example, if you want to make a scatter plot, with a trend line going through it, then setting this field to False for the trend-line component will make the chart snugly fit the points, with the trend line extending beyond.
        :param list error_bars_upper_values: Specifies the tops of error bars. Either a list of values, or if a datasource is specified, an expression in terms of that datasource's variables. If input values is a list, then this must be a list as well.
        :param list error_bars_lower_values: Specifies the bottoms of error bars. Either a list of values, or if a datasource is specified, an expression in terms of that datasource's variables. If input values is a list, then this must be a list as well.
        :param str legend_label: A string label to be used in the legend. If unspecified, this plot will not appear in the legend.
        :param str tooltip_template: Any valid Handlebars template. The template will be provided the x and y values as "x_value" and "y_value", so you can use these to format a tooltip for each point. For example, "<div class="my_awesome_tooltip">When age is {{x_value}}, salary is expected to be {{y_value}}</div>"
        """
        config = {
            'style': style,
            'color': color,
            'line_width': line_width,
            'circle_radius': circle_radius,
            'use_in_plot_range': use_in_plot_range,
            'error_bars_upper_values': error_bars_upper_values,
            'error_bars_lower_values': error_bars_lower_values,
            'legend_label': legend_label,
            'tooltip_template': tooltip_template
        }

        # decide which overload to use
        if isinstance(x, list) and isinstance(y, list):
            return self._plot_data(x, y, config)
        elif isinstance(x, str) and isinstance(y, str):
            return self._plot_expression(x, y, datasource, config)
        else:
            raise Exception("CustomPlot.plot error: x and y inputs must be either lists of numbers or strings referencing a variable or expression")

    def delete(self):
        """
        Delete the Plot and any associated data which has been uploaded.
        """
        super(CustomPlot, self).delete()
        if self.datasource is not None: self.datasource.delete()



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from eureqa.analysis.components.base import _Component
from eureqa.utils.jsonrest import _JsonREST

class ByRowPlot(_Component):
    """
    Represents a line graph for plotting variables by row.

    For example::

       p = ByRowPlot(d, x_var='W', focus_variable='W', plotted_variables=['A'], should_center=True)
       analysis.create_card(p)


    :param eureqa.DataSource datasource: Data source for the component's data
    :param str x_var: Name of the variable to plot as the X axis
    :param list plotted_variables: List of variable names to plot.
    :param str focus_variable: Name of the variable in :attr:`plotted_variables` to bring to the foreground (required)
    :param bool should_center: Should the plot be centered?
    :param bool should_scale: Should the plot scale?

    :var str eureqa.analysis.components.variable_line_graph.ByRowPlot.focus_variable: Focused (foreground) variable for the component.  Must be a member of :attr:`plotted_variables`
    :var str eureqa.analysis.components.variable_line_graph.ByRowPlot.x_var: Name of the variable to plot as the X axis
    :var list eureqa.analysis.components.variable_line_graph.ByRowPlot.plotted_variables: Variables to plot.  (List of string variable names.)
    :var bool eureqa.analysis.components.variable_line_graph.ByRowPlot.should_center: Should the plot be centered?
    :var bool eureqa.analysis.components.variable_line_graph.ByRowPlot.should_scale: Should the plot scale?
    """

    _component_type_str = 'VARIABLE_LINE_GRAPH'

    def __init__(self, datasource=None, x_var=None, plotted_variables=None, focus_variable=None, should_center=None,
                 should_scale=None, _analysis=None, _component_id=None, _component_type=None):
        # Set 'self._options' first; property setters used below populate it
        self._options = {}

        if datasource is not None:
            self.datasource = datasource

        if x_var is not None:
            self.x_var = x_var

        if plotted_variables is not None:
            self.plotted_variables = plotted_variables

        if focus_variable is not None:
            self.focus_variable = focus_variable

        if should_center is not None:
            self.should_center = should_center

        if should_scale is not None:
            self.should_scale = should_scale

        super(ByRowPlot, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)

    @property
    def datasource(self):
        """The data source providing data for this component

        :return: data source providing data for this component
        """
        if hasattr(self, "_eureqa"):
            return self._eureqa.get_data_source_by_id(self._datasource_id)

    @datasource.setter
    def datasource(self, val):
        self._datasource_id = val._data_source_id
        self._eureqa = val._eureqa
        self._update()


    @property
    def focus_variable(self):
        """The variable that is currently in focus (in the foreground) for this component.
        Must be a member of :attr:`plotted_variables`.

        :return: focus_variable for this component
        :rtype: str
        """

        return self._options.get('focusVariable')

    @focus_variable.setter
    def focus_variable(self, val):
        self._options['focusVariable'] = val
        self._update()

    @property
    def x_var(self):
        """The X-axis variable for this component

        :return: the name of the X-axis variable for this component
        :rtype: str
        """

        return self._options.get('xAxisVariable')

    @x_var.setter
    def x_var(self, val):
        self._options['xAxisVariable'] = val
        self._update()

    @property
    def plotted_variables(self):
        """The plotted variables for this component.

        :return: List of the names of the variables being plotted against the X axis
        :rtype: list
        """
        return self._options.get('plottedVariables')

    @plotted_variables.setter
    def plotted_variables(self, val):
        self._options['plottedVariables'] = val
        self._update()

    @property
    def should_center(self):
        """The should_center option for this component.

        :return: whether this plot should be centered
        :rtype: bool
        """

        return self._options.get('shouldCenter')

    @should_center.setter
    def should_center(self, val):
        self._options['shouldCenter'] = val
        self._update()

    @property
    def should_scale(self):
        """The should_scale option for this component.

        :return: whether this plot should be scaled
        :rtype: bool
        """
        return self._options.get('shouldScale')

    @should_scale.setter
    def should_scale(self, val):
        self._options['shouldScale'] = val
        self._update()


    def _fields(self):
        return super(ByRowPlot, self)._fields() + [ 'datasource_id', 'options' ]



#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from parameter_value import ParameterValue

import json

class TopLevelModelParameterValue(ParameterValue):
    """TopLevelModel parameter value for analysis template

    :param Eureqa eureqa: A eureqa connection.
    :param str id: The id of the parameter.
    :param str value: The parameter value.
    :param str datasource_id: The parameter value for the datasource the expression belongs to.
    :param str search_id: The parameter value for the search the expression belongs to.
    :param str solution_id: The parameter value for the solution the expression belongs to.

    :var str id: The id of the parameter.
    :var str TopLevelModelParameterValue.value: The parameter value.
    """

    def __init__(self, eureqa, id, value, datasource_id, search_id, solution_id):
        """Initializes a new instance of the ~TopLevelModelParameter class
        """
        ParameterValue.__init__(self, id, value, 'top_level_model')
        self.datasource_id = datasource_id
        self.search_id = search_id
        self.solution_id = solution_id
        self._eureqa = eureqa

    def get_solution(self):
        if self.datasource_id is not None and self.search_id is not None and self.solution_id is not None:
            data_source = self._eureqa.get_data_source_by_id(self.datasource_id)
            return self._eureqa._get_solution_by_id(data_source._data_source_id, self.search_id, self.solution_id)
        else:
            return None

    def _to_json(self):
        body = {}
        ParameterValue._to_json(self, body)
        if self.datasource_id is not None:
            body["datasource_id"] = self.datasource_id
        if self.search_id is not None:
            body["search_id"] = self.search_id
        if self.solution_id is not None:
            body["solution_id"] = self.solution_id
        return body

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    @staticmethod
    def _from_json(eureqa, body, execution=None):
        param = TopLevelModelParameterValue(eureqa, None, None, None, None, None)
        ParameterValue._from_json(param, body)
        if param._type != "top_level_model": raise Exception("Invalid type '%s' specified for top_level_model parameter value" % param._type)
        if "datasource_id" in body and "search_id" in body and "solution_id" in body:
            param.datasource_id = body["datasource_id"]
            param.search_id = body["search_id"]
            param.solution_id = body["solution_id"]
        return param



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from eureqa.analysis.components.base import _Component
from eureqa.analysis.analysis_file import AnalysisFile

class DownloadFile(_Component):
    """
    DownloadFile component
    Represents a Download link for a file within an Analysis, including the contents of that file.

    For example::

        df = DownloadFile(file_content="The file content", link_text="The link Text", filename="download.txt")
        analysis.create_html_card('Download file here: {0}'.format(analysis.html_ref(df)))


    :param str file_content: The contents of the file as a str() or eureqa.analysis.analysis_file.AnalysisFile
    :param str link_text: The text of the HTML download link in the Analysis
    :param str filename: The name of the downloaded file
    """

    _component_type_str = 'DOWNLOAD_FILE'

    def __init__(self, file_content=None, link_text=None, filename=None,
                 _analysis=None, _component_id=None, _component_type=None):
        if link_text is not None:
            self._link_text = link_text

        self._file_content = file_content
        self._filename = filename

        super(DownloadFile, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)

    @property
    def link_text(self):
        """
        Text of the link as rendered by the component
        """
        return getattr(self, "_link_text", None)

    @link_text.setter
    def link_text(self, val):
        self._link_text = val
        self._update()

    @property
    def file(self):
        """
        :rtype: eureqa.analysis.analysis_file.AnalysisFile
        """
        if hasattr(self, "_file_id") and hasattr(self, "_analysis") and not hasattr(self, "_file"):
            self._file = AnalysisFile(self._analysis, self._file_id)
        return getattr(self, "_file", None)

    def _upload_file(self, analysis):
        # Associate with the provided AnalysisFile
        if isinstance(self._file_content, AnalysisFile):
            self._file = self._file_content
            self._file_id = self._file._file_id
        # Or, if that doesn't exist, make one
        elif self._file_content is not None:
            file = self._file_content
            self._file = AnalysisFile.create(analysis, file, filename=self._filename)
            self._file_id = self._file._file_id

    def _pre_associate(self, analysis):
        self._upload_file(analysis)

    def _associate_with_table(self, analysis):
        self._upload_file(analysis)

        # ideally, file_url should be returned by the app server, but if we do a
        # POST to the app server here, it requires another network operation and
        # the component goes into analysis' main component list, which isn't
        # desired for components used in table
        self._file_url = ("/api/%s/analysis/%s/files/%s" % (analysis._organization, analysis._id, self._file_id))
        return super(DownloadFile, self)._associate_with_table(analysis)


    def _fields(self):
        return super(DownloadFile, self)._fields() + [ "link_text", "file_id", "file_url" ]



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime
import getpass
import json
import os
import os.path
import sys

class _CredentialGetter:
    @staticmethod
    def _get_password_file():
        # store password file outside of the source directory
        PASSWORD_FILE = '.eureqa_passwd'
        path = os.path.expanduser("~")
        return path + '/' + PASSWORD_FILE

    @classmethod
    def _clear_passwords(cls):
        try: os.remove(cls._get_password_file())
        except: return

    @classmethod
    def _load_passwords(cls):
        try:
            with open(cls._get_password_file(), 'rb') as f:
                return json.load(f)
        except:
            return {}

    @staticmethod
    def login_timestamp(login):
        try:
            return datetime.datetime(login['last_used'])
        except:
            return datetime.datetime(2000, 1, 1)

    @classmethod
    def _find_login(cls, data, url, username=None):
        if url not in data: return None

        max_date = None
        max_date_i = None
        for i, login in enumerate(data[url]):
            if username and login['username'] != username:
                continue
            date = cls.login_timestamp(login)
            if not max_date or date > max_date:
                max_date = date
                max_date_i = i
        return max_date_i

    @staticmethod
    def _prompt_username():
        username = raw_input('ENTER USERNAME: ')
        if not username:
            raise Exception('Login cancelled by user')
        return username

    @staticmethod
    def _prompt_password_spyder():
        return raw_input('ENTER PASSWORD: ').encode('base64')

    @staticmethod
    def _prompt_password():
        if any('SPYDER' in name for name in os.environ) or 'pythonw.exe' in sys.executable:
            password = _prompt_password_spyder()
        else:
            password = getpass.getpass('ENTER PASSWORD: ').encode('base64')
        if not password: raise Exception('Login cancelled by user')
        return password

    ### MAIN FUNCTIONALITY: ###

    @staticmethod
    def prompt_one_time_password(login):
        password = raw_input('ENTER ONE_TIME PASSWORD SENT BY SMS: ')
        if not password: raise Exception('Login cancelled by user')
        login['one_time_password'] = password
        return login

    @classmethod
    def store_login(cls, url, login):
        """Saves login credentials to they are used automatically next time.
        Credentials are saved to a file your home directory.

        :param str url: URL to login destination.
        :param dict login: Login credentials as dictionary.
        """
        if login.has_key('one_time_password'): del login['one_time_password']
        data = cls._load_passwords()
        if url not in data: data[url] = []
        login['last_used'] = str(datetime.datetime.now())
        index = cls._find_login(data, url, login['username'])
        if index is None:
            data[url].append(login)
        else:
            data[url][index] = login
        with open(cls._get_password_file(), 'wb') as f:
            json.dump(data, f, indent=4, default=str)

    @classmethod
    def get_login(cls, url, username=None, password=None, key=None, force_prompt=False):
        """Returns (username, password) for a Eureqa SaaS url and prompts user for
        credentials if not found. Passwords should be base64 encoded.

        :param str url: URL for login connection.
        :param str username: Login username or most recently used if not provided.
        :param str password: Password for login.
        :param str key: Access key for login.  This is the preferred method.
        :param bool force_prompt: Whether the user will always be prompted to enter credentials.

        """
        data = cls._load_passwords()

        index = cls._find_login(data, url, username)
        login = data[url][index] if index is not None and (username is None or password is None) else {'username': username, 'password': password}
        if key:
            login['key'] = key

        ask_username = force_prompt or not login['username']
        ask_password = (force_prompt or not login['password']) and not key

        if ask_username or ask_password: print('PLEASE LOGIN,')
        if ask_username:
            login['username'] = cls._prompt_username()
        elif ask_password:
            print('LOGGING IN WITH USERNAME: ' + login['username'])
        if ask_password: login['password'] = cls._prompt_password()

        if login.has_key('one_time_password'): del login['one_time_password']

        return login




# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from eureqa.analysis.components import *
from eureqa.analysis.cards import Card
from eureqa.data_source import DataSource
from eureqa.analysis.analysis_file import AnalysisFile

import warnings

class Analysis(object):
    """Represents an analysis on the server.

    :var str `~eureqa.analysis.Analysis.name`: The analysis name.
    :var str `~eureqa.analysis.Analysis.description`: The analysis description.
    """

    def __init__(self, body, eureqa):
        """For internal use only
        PARAM_NOT_EXTERNALLY_DOCUMENTED
        """

        self._id = body['analysis_id']
        self._owner = body['owner']
        self._eureqa = eureqa
        if not 'description' in body:
            body['description'] = ''
        self._body = body

    def url(self):
        return '%s/%s/analyses/%s' % (self._eureqa._session.url, self._eureqa._session.organization, self._id)

    @property
    def _organization(self):
        return self._eureqa._session.organization

    @property
    def name(self):
        """The name of the analysis"""

        return self._body['analysis_name']

    def __repr__(self):
        return "Analysis(" + repr(self.name) + ", ...)"

    @property
    def analysis_id(self):
        """The id of the analysis"""

        return self._id

    @name.setter
    def name(self, value):
        """Change the analysis name to the new value.

        :param str value: The new value for the analysis name.
        """

        self._body['analysis_name'] = value
        self._update()

    @property
    def description(self):
        """The description of the analysis"""

        return self._body['description']

    @description.setter
    def description(self, value):
        """Change the analysis description to the new value.

        :param str value: The new value for the analysis description.
        """

        self._body['description'] = value
        self._update()

    def delete(self):
        """Deletes the analysis from the server."""

        endpoint = '/analysis/%s' % self._id
        self._eureqa._session.report_progress('Deleting analysis: \'%s\'.' % self._id)
        self._eureqa._session.execute(endpoint, 'DELETE')

    def get_cards(self):
        """ Get all Analysis Cards associated with this Analysis

        :return: A list of Card objects that are rendered as part of this Analysis
        :rtype: list of :class:`~eureqa.analysis.cards.Card`
        """
        # Inline import to avoid circular import (Item imports Analysis)
        from .cards import Card
        return sorted(Card._get_all(self), key = lambda c: c._order_index)

    def get_components(self):
        """ Get all Analysis Components associated with this Analysis

        :return: A list of Component objects attached to this Analysis, that may be rendered by Items in the Analysis
        :rtype: list of :class:`~eureqa.analysis.components._Component`
        """
        # Inline import to avoid circular import (Component imports Analysis)
        from .components import _Component
        c = _Component()
        c._analysis = self
        c._eureqa = self._eureqa
        return c._get_all_from_self()

    def _get_component_by_id(self, component_id):
        """ Internal: return the Component object from this analysis with the specified id """
        from .components import _Component
        return _Component._get(_analysis=self, _component_id=component_id)

    def _create_card(self, title, description, collapse, component):

        # The "create_card_item" endpoint isn't quite clever enough to
        # handle Image cards on its own.  So give it a little help.
        if isinstance(component, Image):
            component._upload_image(self)
        
        body = {
            "title": title,
            "description": description,
            "collapse": collapse,
            "component": component._to_json()
        }
        endpoint = '/analysis/%s/items/create_card_item' % self._id
        results = self._eureqa._session.execute(endpoint, 'POST', body)

        # Split the JSON bundle into a tree of Components and a root Item
        components = {}
        for component_json in results["components"]:
            c = _Component._construct_from_json(component_json, _analysis=self)
            components[c._component_id] = c

        # If we have a Component that contains another Component,
        # set up the corresponding Python reference.
        for c in components.itervalues():
            if isinstance(c, TitledLayout):
                c.create_card(components[c._content_component_id])

        card = Card()
        card._from_json(results["item"])
        card.component = components[card._default_component_id]

        return card

    def create_model_card(self, solution, title=None, description=None, collapse=False):
        """Creates a new model card.  Adds the card to this analysis.

        :param eureqa.Solution solution: The solution that will be displayed on the card.
        :param str title: The card title.
        :param str description: The card description.
        :param bool collapse: Whether the card should default to be collapsed.
        :return: An object that represents the newly created card.
        :rtype: ~eureqa.analysis_cards.ModelCard
        """
        component = Model(solution=solution)
        return self._create_card(title, description, collapse, component)

    def create_model_summary_card(self, solution, collapse=False):
        """Creates a new model summary card.  Adds the card to this analysis.

        :param eureqa.Solution solution: The solution that will be displayed on the card.
        :param bool collapse: Whether the card should default to be collapsed.
        :return: An object that represents the newly created card.
        :rtype: ~eureqa.analysis_cards.ModelSummaryCard
        """
        # ModelSummaryCards do not have a title or description.
        # So don't use our handy wrapper which automatically creates a TitleLayout.
        from .cards import Card
        component = ModelSummary(_analysis=self, solution=solution)
        return Card(_analysis=self, _component=component)


    def create_distribution_plot_card(self, data_source, variable, title=None, description=None, collapse=False):
        """Creates a new distribution plot card.  Adds the card to this analysis.

        :param DataSource data_source: The data source to which the variable belongs.
        :param str variable: The name of the variable that will be displayed on the card.
        :param str title: The card title.
        :param str description: The card's description.
        :param bool collapse: Whether the card should default to be collapsed.
        :return: An object that represents the newly created card.
        :rtype: ~eureqa.analysis_cards.DistributionPlotCard
        """
        component = DistributionPlot(data_source, variable)
        return self._create_card(title, description, collapse, component)

    def create_text_card(self, text, title='Text', description=None, collapse=False):
        """Creates a new text card.  Adds the card to this analysis.

        :param str text: The card text.
        :param str title: The card title.
        :param str description: Deprecated; unused.  (This card doesn't have a Description.)
        :param bool collapse: Whether the card should default to be collapsed.
        :return: An object that represents the newly created card.
        :rtype: ~eureqa.analysis_cards.TextCard
        """
        component = TextBlock(text)
        return self._create_card(title, description, collapse, component)

    def create_html_card(self, html, title='HTML', description=None, collapse=False):
        """Creates a new HTML card.  Adds the card to this analysis.

        :param str html: The card's HTML body.
        :param str title: The card title.
        :param str description: Deprecated; unused.  (This card doesn't have a Description.)
        :param bool collapse: Whether the card should default to be collapsed.
        :return: An object that represents the newly created card.
        :rtype: ~eureqa.analysis_cards.HtmlCard
        """
        component = HtmlBlock(html)
        return self._create_card(title, description, collapse, component)

    def create_custom_plot_card(self, plot, title=None, description=None, collapse=False):
        """Creates a new custom plot card.  Adds the card to this analysis.

        :param ~eureqa.analysis.components.CustomPlot plot: The Plot to be displayed in the card.
        :param str title: The card title.
        :param str description: The card's description.
        :param bool collapse: Whether the card should default to be collapsed.
        :return: An object that represents the newly created card.
        :rtype: Card
        """

        # First, upload data for the plot
        plot.upload_data(self._eureqa)

        # Then add the plot to the analysis
        return self._create_card(title, description, collapse, plot)

    def create_model_evaluator_card(self, solution, title=None, description=None, collapse=False,
                                    *args, **kwargs): #*args and **kwargs - for backward compatibility
        """Creates a new model evaluator card.  Adds the card to this analysis.

        :param eureqa.Solution solution: Solution to fetch results from for the primary model evaluator
        :param str title: Title of the card.  Defaults to 'Evaluate Model'.
        :param str description: Description of the card.
        :param bool collapse: Whether the card should default to be collapsed.
        :return: Object representing the created card.
        :rtype: Card

        Additional models can be added to this model by calling the
        :class:`~eureqa.analysis.ModelEvaluatorCard.add_solution_info` method on
        the returned object.
        """

        if isinstance(solution, DataSource):
            # Backward compatibility support for
            # (self, datasource, solution, title, description, collapse)
            # calling convention.
            solution = title
            title = description

            # collapse's default value is False instead of None
            # It will cause description to be wrong if it is moved without being set
            # collapse should only be moved to description if it is a string
            if isinstance(collapse, str):
                description = collapse
                collapse = False
            else:
                description = None

            if len(args) >= 1:
                collapse = args[0]


        if title is None:
            title = "Evaluate Model"

        component = ModelEvaluator()
        component.add_solution_info(solution)
        return self._create_card(title, description, collapse, component)

    def create_box_plot_card(self, datasource, x_var, y_var, title=None, description=None, needs_guides = False, axis_labels=None, label_format=None, collapse=False):
        """Creates a new box-plot card.  Adds the card to this analysis.

        :param DataSource datasource: Data source for the card's data
        :param str x_var: The X-axis variable for the card's plot
        :param str y_var: The Y-axis variable for the card's plot
        :param str title: The title of the card
        :param str description: The description of the card
        :param bool needs_guides: Whether the card needs guides
        :param list axis_labels: Axis labels for this card's plot
        :param str label_format: Label format for this card
        :param bool collapse: Whether the card should default to be collapsed.
        :return: Object representing the created card
        :rtype: ~eureqa.analysis_cards.BoxPlotCard
        """
        return self._create_two_variable_plot_card(BoxPlot, datasource, x_var, y_var, title, description,
                                                   needs_guides, axis_labels, label_format, "box plot", collapse)

    def create_double_histogram_plot_card(self, datasource, x_var, y_var, title=None, description=None, needs_guides=False,
                                          axis_labels=None, label_format=None, collapse=False):
        """Creates a new double-histogram card.  Adds the card to this analysis.

        :param DataSource datasource: Data source for the card's data
        :param str x_var: The X-axis variable for the card's plot
        :param str y_var: The Y-axis variable for the card's plot
        :param str title: The title of the card
        :param str description: The description of the card
        :param bool needs_guides: Whether the card needs guides
        :param list axis_labels: Axis labels for this card's plot
        :param str label_format: Label format for this card
        :param bool collapse: Whether the card should default to be collapsed.
        :return: Object representing the created card
        :rtype: ~eureqa.analysis_cards.DoubleHistogramPlotCard
        """

        return self._create_two_variable_plot_card(DoubleHistogramPlot, datasource, x_var, y_var, title, description,
                                                   needs_guides, axis_labels, label_format, "histogram plot", collapse)

    def create_scatter_plot_card(self, datasource, x_var, y_var, title=None, description=None, needs_guides=False,
                                 axis_labels=None, label_format=None, collapse=False):
        """Creates a new scatter-plot card.  Adds the card to this analysis.

        :param DataSource datasource: Data source for the card's data
        :param str x_var: The X-axis variable for the card's plot
        :param str y_var: The Y-axis variable for the card's plot
        :param str title: The title of the card
        :param str description: The card's description
        :param bool needs_guides: Whether the card needs guides
        :param list axis_labels: Axis labels for this card's plot
        :param list label_format: Label format for this card
        :param bool collapse: Whether the card should default to be collapsed.
        :return: Object representing the created card
        :rtype: ~eureqa.analysis_cards.ScatterPlotCard
        """

        return self._create_two_variable_plot_card(ScatterPlot, datasource, x_var, y_var, title, description,
                                                   needs_guides, axis_labels, label_format, "scatter plot", collapse)

    def create_binned_mean_plot_card(self, datasource, x_var, y_var, title=None, description=None, needs_guides=False,
                                     axis_labels=None, label_format=None, collapse=False):
        """Creates a new binned-mean-plot card.  Adds the card to this analysis.

        :param ~DataSource datasource: Data source for the card's data
        :param str x_var: The X-axis variable for the card's plot
        :param str y_var: The Y-axis variable for the card's plot
        :param str title: The title of the card
        :param str description: The card's description.
        :param bool needs_guides: Whether the card needs guides
        :param list axis_labels: Axis labels for this card's plot
        :param list label_format: Label format for this card
        :param bool collapse: Whether the card should default to be collapsed.
        :return: Object representing the created card
        :rtype: ~eureqa.analysis_cards.BinnedMeanPlotCard
        """

        return self._create_two_variable_plot_card(BinnedMeanPlot, datasource, x_var, y_var, title, description,
                                                   needs_guides, axis_labels, label_format, "scatter plot", collapse)

    def _create_two_variable_plot_card(self, plot_type, datasource, x_var, y_var, title=None, description=None,
                                       needs_guides=False, axis_labels=None, label_format=None, card_name="two-variable", collapse=False):

        component = plot_type(datasource=datasource,
                              axis_labels=axis_labels,
                              label_format=label_format,
                              needs_guides=needs_guides,
                              x_var=x_var,
                              y_var=y_var)

        return self._create_card(title, description, collapse, component)

    def create_by_row_plot_card(self, datasource, x_var, plotted_vars, title=None, description=None,
                                focus_variable=None, should_center=True, should_scale=False, collapse=False):
        """Create a new by-row plot card.  Adds the card to this analysis.

        :param DataSource datasource: Data source for the card's data
        :param str x_var: Name of the variable to plot as the X axis
        :param str plotted_vars: List of string-names of variables to plot.
               (To modify a variable's display name, first create the card; then modify the display name directly on it.)
        :param str title: The card's title.
        :param str description: The card's description.
        :param str focus_variable: Name of the variable in 'plotted_vars' to bring to the foreground
        :param bool should_center: Should the plot be centered?
        :param bool should_scale: Should the plot scale?
        :param bool collapse: Whether the card should default to be collapsed.
        :return: Object representing the created card
        :rtype: ~eureqa.analysis_cards.ByRowPlotCard
        """

        for x in plotted_vars:
            assert isinstance(x, basestring), "'plotted_vars' must be a list of variable-name strings"
        if (not focus_variable) and len(plotted_vars) > 0:
            focus_variable = plotted_vars[0]

        component = ByRowPlot(datasource, x_var, plotted_vars,
                              focus_variable, should_center, should_scale)

        return self._create_card(title, description, collapse, component)

    def create_model_fit_by_row_plot_card(self, solution, title=None, description=None, collapse=False, x_axis=None):
        """Create a new model fit by-row plot card.  Adds the card to this analysis.
        Note that by-row plots are meant for use with Numeric searches.  They may not work
        properly if used with other types of searches.

        :param ~eureqa.Solution solution: The Solution object for which this card is being created
        :param str title: The card's title.
        :param str description: The card's description.
        :param bool collapse: Whether the card should default to be collapsed.
        :param str x_axis: **Ignored**
        :return: Object representing the created card
        :rtype: ~eureqa.analysis_cards.ModelFitByRowPlotCard
        """
        if x_axis is not None:
            warnings.warn("Axis labels are no longer supported by ModelFitByRow plots", DeprecationWarning)

        component = ModelFitByRowPlot(solution=solution)
        return self._create_card(title, description, collapse, component)

    def create_model_fit_separation_plot_card(self, solution, title=None, description=None, collapse=False):
        """Create a new model fit separation-plot card.  Adds the card to this analysis.
        Note that separation plots are meant for use with classification searches.  They may not work
        properly if used with other types of searches.

        :param ~eureqa.Solution solution: The Solution object for which this card is being created
        :param str title: The card's title.
        :param str description: The card's description.
        :param bool collapse: Whether the card should default to be collapsed.
        :return: Object representing the created card
        :rtype: ~eureqa.analysis_cards.ModelFitSeparationPlotCard
        """

        component = ModelFitSeparationPlot(solution=solution)
        return self._create_card(title, description, collapse, component)

    def _get_card_type_from_search_type(self, search_type):
        card_types = {
            "generic": ModelFitByRowPlot,
            "timeseries": ModelFitByRowPlot,
            "classification": ModelFitSeparationPlot,
            "timeseries_classification": ModelFitSeparationPlot
            }
        return card_types[search_type]

    def _get_card_typename_from_search_type(self, search_type):
        card_typenames = {
            "generic": "model fit by row plot",
            "timeseries": "model fit by row plot",
            "classification": "model fit separation plot",
            "timeseries_classification": "model fit separation plot"
            }
        return card_typenames[search_type]

    def create_model_fit_plot_card(self, solution, x_axis=None, title=None, description=None, collapse=False):
        """Create a new model fit card.  Adds the card to this analysis.
        Automatically choose the correct type of card (by-row plot or separation plot) based on
        the specified search.  Numeric searches use by-row plots; time-series searches use separation plots.

        :param ~eureqa.Solution solution: The Solution object for which this card is being created
        :param str x_axis: **Ignored**
        :param str title: The card's title.
        :param str description: The card's description.
        :param bool collapse: Whether the card should default to be collapsed.
        :return: Object representing the created card
        :rtype: ~eureqa.analysis_cards.model_fit_plot_card.ModelFitPlotCard
        """

        if x_axis is not None:
            warnings.warn("Axis labels are no longer supported by Model Fit plots", DeprecationWarning)

        search_type = solution.search._body['search_template_id'].lower()
        card_type = self._get_card_type_from_search_type(search_type)

        component = card_type(solution=solution)
        return self._create_card(title, description, collapse, component)


    def add_card(self, card):
        """ Add an analysis Item to this analysis

        :param Card card: Card to add to the analysis. Must not have been previously added or retrieved from an Analysis.
        """
        assert getattr(card, "_analysis", self) == self, \
            "'card' must not already be associated with another analysis.  (Did you mean to pass in <card>.clone()?)"
        card._associate(self)

    def create_card(self, component):
        """ Create a new Card on this Analysis containing only the specified Component

        :param Component component: instance of the class add to the created Item
        :return: the created :class:Card

        Example::

            card = e.create_card(comp)

        """
        from .cards import Card
        card = Card(_component = component)
        self.add_card(card)
        return card

    def html_ref(self, component):
        component._associate(self)

        # Technically you can either add this HtmlBlock, or you can pass in the "analysis"
        # constructor argument back when you created it, or you can add `component` now and add this HtmlBlock later.
        # But if you don't already know that, then you almost certainly just want to add this HtmlBlock.
        assert hasattr(component, "_component_id"), \
            "Add this HtmlBlock to an Analysis (or to another Component that is already on an Analysis) before adding other Components to it"

        return '<nutonian-component id="%s" />' % component._component_id

    def upload_image(self, image_path):
        """Upload an image to the server, to be embedded in analysis cards.

        :param str image_path: the filepath to the image on your filesystem.
        :return: An object that represents the newly uploaded image.
        :rtype: ~eureqa.analysis.components.Image
        """

        image = Image(image_path)
        image._associate(self)
        return image

    def create_image_card(self, image_path, title=None, description=None, collapse=False):
        """Creates a new text card containing only header text and one image.

        :param str image_path: the filepath to the image in your filesystem.
        :param str title: The card title.
        :param str description: a description of the card, to appear above the image
        :param bool collapse: Whether the card should default to be collapsed.
        :return: An object that represents the newly created card.
        :rtype: ~eureqa.analysis_cards.TextCard
        """

        if title is None: title = 'Image'

        component = Image(image_path)
        return self._create_card(title, description, collapse, component)

    def _create_file(self, filename, file):

        # If the file already exists, delete it
        try:
            self._get_file(filename).delete()
        except:
            pass

        return AnalysisFile.create(self, file, filename)

    def _get_file(self, filename):
        files = AnalysisFile._get_for_analysis(self)
        files_with_name = [x for x in files if x._filename == filename]

        if len(files_with_name) == 0:
            raise Exception("File not found")

        if len(files_with_name) > 1:
            warnings.warn("Multiple files found with the name '%s'.  Picking one.  " % (filename))

        # Maybe this will be the most-recent file?
        return files_with_name[-1]

    def _update(self):
        endpoint = '/analysis/%s' % self._id
        self._eureqa._session.report_progress('Updating analysis: \'%s\'.' % self._id)
        new_body = self._eureqa._session.execute(endpoint, 'PUT', self._body)
        new_instance = Analysis(new_body, self._eureqa)
        self.__dict__ = new_instance.__dict__



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from eureqa.analysis.components.base import _Component
from eureqa.data_source import DataSource
import warnings

class ModelEvaluator(_Component):
    """This component evaluates models against different datasources and
    compares their performance.

    Additional models can be added to this model by calling
    :meth:`add_solution_info`

    For example::

        c = ModelEvaluator(solutions=s.get_solutions(), datasource=d, search=s, solution=s.get_best_solution())
        analysis.create_card(c)



    :param list[Solution] solutions: Solutions to evaluate against the data
    :param Solution solution: Solution to fetch results from for the primary model evaluator

    """

    _component_type_str = 'EVALUATE_MODEL'

    def __init__(self, solutions=None, solution=None,
                 _analysis=None, _component_id=None, _component_type=None,
                 **kwargs): #**kwargs - for backward compatibility
        if isinstance(solution, DataSource):
            # This is backward compatibility hack to support the
            # (self, solutions, datasource, search, solution) calling signature.
            solution = _component_id

            # Only support backwared compatible signature for public params
            # because the internal calls were ported to new signature.
            _analysis = None
            _component_id = None
            _component_type = None

        # Field is required by the backend model
        self._evaluationInfo = []

        if solution is not None:
            self.solution = solution

        if solutions:
            for soln in solutions:
                self.add_solution_info(soln)

        super(ModelEvaluator, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)


    class SolutionInfo(object):
        """**Do not instantiate directly**. Use :meth:`ModelEvaluator.add_solution_info`
        instead. The solution information for a single model
        evaluation ("tab") on ModelEvaluatorCard.

        :param str body: internal
        :param _Component component: internal

        :var str ~eureqa.analysis_cards.ModelEvaluatorCard.SolutionInfo.datasource_id: ID of the DataSource referenced by this solution-tab
        :var str ~eureqa.analysis_cards.ModelEvaluatorCard.SolutionInfo.search_id: ID of the Search referenced by this solution-tab
        :var str ~eureqa.analysis_cards.ModelEvaluatorCard.SolutionInfo.solution_id: ID of the Solution referenced by this solution-tab
        :var bool ~eureqa.analysis_cards.ModelEvaluatorCard.SolutionInfo.has_target_variable: Whether the datasource contains the target variable

        """

        def __init__(self, component, body):
            """ SolutionInfo init """
            if hasattr(component, "_analysis"):
                self._eureqa = component._analysis._eureqa
            elif hasattr(component, "_eureqa"):
                self._eureqa = component._eureqa
            self._component = component
            self._from_json(body)

        @classmethod
        def _from_solution(cls, component, solution):
            instance = cls(
                component=component,
                body={
                    "datasource_id": solution._datasource_id,
                    "search_id": solution._search_id,
                    "solution_id": solution._id
                })
            return instance

        def _to_json(self):
            self._ensure_target_variable()
            return self._body

        def _from_json(self, body):
            self._body = body

            # Value is computed in Python
            if "hasTargetVariable" in self._body:
                del body["hasTargetVariable"]

        @property
        def solution(self):
            """The Solution that is being explained

            :rtype: Solution
            """
            datasource_id = self._body.get("datasource_id")
            search_id = self._body.get("search_id")
            solution_id = self._body.get("solution_id")
            if datasource_id and search_id and solution_id:
                return self._eureqa._get_solution_by_id(datasource_id, search_id, solution_id)

        @solution.setter
        def solution(self, val):
            self._body["solution_id"] = val._id
            self._body["datasource_id"] = val._datasource_id
            self._body["search_id"] = val._datasource_id
            self._eureqa = val._eureqa
            self._update()

        def _ensure_target_variable(self):
            # If we don't have a cached computed value for this variable,
            # and we have enough fields to compute it,
            # then go compute it.
            if "hasTargetVariable" not in self._body:
                solution = self.solution
                datasource = solution.search.get_data_source()
                target_variable = datasource.get_variable_details(solution.target)
                self._body['hasTargetVariable'] = target_variable is not None and target_variable.distinct_values > 0

        @property
        def has_target_variable(self):
            """Whether the datasource contains the solution's target variable.

            If the datasource contains the target variable, the standard plot of this
            Component may include the raw target-variable data for comparison alongside
            the computed value.

            :return: Whether the datasource contains the target variable
            """
            self._ensure_target_variable()
            return self._body.get('hasTargetVariable')

        @has_target_variable.setter
        def has_target_variable(self, val):
            self._body['hasTargetVariable'] = val
            self._update()

        @property
        def accuracy(self):
            """
            Accuracy of this Solution.  Rendered as a human-readable pretty-printed string.

            :rtype: str
            """
            return self._body.get("accuracy")

        @accuracy.setter
        def accuracy(self, val):
            self._body["accuracy"] = val
            self._update()

        @property
        def is_fetching(self):
            return self._body.get("isFetching")

        @is_fetching.setter
        def is_fetching(self, val):
            self._body["isFetching"] = val
            self._update()

        def _update(self):
            self._component._update()
            self._from_json(self._body)

    def add_solution_info(self, solution, *args):
        """Add a new (non-default) solution and tab to this card.
        Once added, this object will show up in the list returned by `solution_infos`.

        :param Solution solution: Solution associated with the model evaluation.
        """

        if isinstance(solution, DataSource):
            # This is backward compatibility hack to support the
            # (self, datasource, solution) calling signature.
            solution = args[0]

        self._eureqa = solution._eureqa

        solution_info = self.SolutionInfo._from_solution(self, solution)

        if not hasattr(self, "_originalSolutionInfo"):
            # First submission gets added as the original solution info
            self._originalSolutionInfo = solution_info._to_json()
            return

        self._evaluationInfo.append(solution_info._to_json())

        self._eureqa = solution._eureqa
        self._update()

    def clear_solution_infos(self):
        """
        Remove all existing solution infos from the current Component
        """
        del self._originalSolutionInfo
        self._evaluationInfo = []

    @property
    def solution_infos(self):
        """The set of all SolutionInfo objects associated with this card.
        One per solution tab displayed in the UI.
        Note that `solution_infos[0]` is the default card; it may be treated specially by the UI.

        To add a solution, use the "add_solution_info()" method.

        :return: List or tuple of :class:`SolutionInfo` objects
        """
        # Pass in our datasource/search/solution in case they are the same as that of the new SolutionInfo.
        # Saves an RPC round-trip for each if so.
        # SolutionInfo is responsible for figuring out that we passed in the wrong objects and silently ignoring them.
        return tuple(self.SolutionInfo(self, x) for x in
                     (([self._originalSolutionInfo] if hasattr(self, "_originalSolutionInfo") else []) +
                      self._evaluationInfo))
    # No setter.  Mutate the returned SolutionInfo objects or use the "add_solution_info" method.


    @property
    def solution(self):
        """The Solution that is being explained

        :rtype: Solution
        """
        if hasattr(self, "_eureqa"):
            return self._eureqa._get_solution_by_id(self._datasource_id, self._search_id, self._solution_id)

    @solution.setter
    def solution(self, val):
        self._solution_id = val._id
        self._search_id = val._search_id
        self._datasource_id = val._datasource_id
        self._eureqa = val._eureqa
        self._update()

    def _fields(self):
        return super(ModelEvaluator, self)._fields() + [ 'evaluationInfo', 'originalSolutionInfo', 'datasource_id', 'search_id', 'solution_id' ]



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import missing_value_policies


class VariableOptions(object):
    """Represents a set of settings for a variable that can be included into a search. Overrides any default behavior for
    the specified variable. 

    :param str name: The name of the variable. It is the same as a column name in the list of data source columns.
    :param str missing_value_policy: A policy name for the processing of missing values. One of the values from missing_value_policies module can be used.
    :param bool remove_outliers_enabled: Enables removal of rows which contain outlier.
    :param float outlier_threshold: Controls the threshold for whether a point is considered an outlier. A higher value means points must be further from the mean to be considered outliers.
    :param bool normalize_enabled: Enables normalization of the variable.
    :param int normalization_offset: A constant offset to subtract from each value.
    :param int normalization_scale: A constant factor to divide each value by before adding the offset.
    :param int min_delay: The number of time increments into the future to predict with a time series search.

    :var str ~VariableOptions.name: The name of the variable. It is the same as a column name in the list of data source columns.
    :var bool missing_values_enabled: Enables additional processing of missing values.
    :var str missing_value_policy: A policy name for the processing of missing values. One of the values from
        missing_value_policies module can be used. missing_value_policies.column_mean is used if the additional
        processing of missing values is enabled but no policy name provided.
    :var float outlier_threshold: The threshold at which a point is considered an outlier and will be removed.
    """

    def __init__(self, name=None, missing_value_policy=missing_value_policies.column_iqm,
                 remove_outliers_enabled=False, outlier_threshold=None, normalize_enabled=False,
                 normalization_offset=None, normalization_scale=None, min_delay=None, _update_parent_settings=None):

        self.missing_value_policy = missing_value_policy

        self.remove_outliers_enabled = remove_outliers_enabled
        if remove_outliers_enabled and outlier_threshold is None:
            self.outlier_threshold = 2
        else:
            self.outlier_threshold = outlier_threshold

        self.normalize_enabled = normalize_enabled
        if normalize_enabled and normalization_offset is None:
            self.normalization_offset = '<none>'
        else:
            self.normalization_offset = normalization_offset

        if normalize_enabled and normalization_scale is None:
            self.normalization_scale = '<none>'
        else:
            self.normalization_scale = normalization_scale

        self._min_delay = min_delay

        self.name = name

        self._filter_enabled = False
        self._filter_expression = None

        self._update_parent_settings = _update_parent_settings


    @property
    def min_delay(self):
        return self._min_delay
    @min_delay.setter
    def min_delay(self, val):
        self._min_delay = val
        if self._update_parent_settings is not None:
            self._update_parent_settings()

    def _set_filter(self, filter_enabled, filter_expression):
        # It is more appropriate to expose filter properties as a search options, not variable options.
        # Therefore they are not exposed as public fields. But we have to keep them at this level
        # because that's what the backend expects.
        self._filter_enabled = filter_enabled
        self._filter_expression = filter_expression

    def _to_json(self):
        body = {'variable_name': self.name}

        if self._filter_expression:
            body['filter_enabled'] = True
            body['filter_expression'] = self._filter_expression
        else:
            body['filter_enabled'] = False

        if self.missing_value_policy:
            body['missing_value_policy'] = self.missing_value_policy

        if self.outlier_threshold:
            body['remove_outliers_enabled'] = True
            body['outlier_threshold'] = self.outlier_threshold
        else:
            body['remove_outliers_enabled'] = False

        if self.normalization_offset:
            body['normalize_enabled'] = True
            body['normalize_offset'] = self.normalization_offset
            body['normalize_scale'] = self.normalization_scale
        else:
            body['normalize_enabled'] = False

        if self.min_delay is not None:
            body['min_delay'] = self.min_delay
        return body

    def __from_json(self, body):
        self.name = body['variable_name']
        self.missing_value_policy = body.get('missing_value_policy')
        self.remove_outliers_enabled = body.get('remove_outliers_enabled')
        self.outlier_threshold = body.get('outlier_threshold')
        self.normalize_enabled = body.get('normalize_enabled')
        self.normalization_offset = body.get('normalize_offset')
        self.normalization_scale = body.get('normalize_scale')
        self._filter_enabled = body.get('filter_enabled')
        self._filter_expression = body.get('filter_expression')
        self._min_delay = body.get('min_delay')
        self._body = body

    @classmethod
    def _from_json(cls, body, _update_parent_settings=None):
        variableOptions = VariableOptions(_update_parent_settings=_update_parent_settings)
        variableOptions.__from_json(body)
        return variableOptions

    def __eq__(self, other):
        if not isinstance(other, VariableOptions):
            return False
        return (self.name == other.name
                and self.missing_value_policy == other.missing_value_policy
                and self.remove_outliers_enabled == other.remove_outliers_enabled
                and self.outlier_threshold == other.outlier_threshold
                and self.normalize_enabled == other.normalize_enabled
                and self.normalization_offset == other.normalization_offset
                and self.normalization_scale == other.normalization_scale
                and self._filter_enabled == other._filter_enabled
                and self._filter_expression == other._filter_expression
                and self.min_delay == other.min_delay)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        tpl = (self.missing_value_policy, self.remove_outliers_enabled,
               self.outlier_threshold, self.normalize_enabled, self.normalization_offset, self.normalization_scale,
               self._filter_enabled, self._filter_expression, self.name, self.min_delay)
        return hash(tpl)

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)



from eureqa.analysis.components.base import _Component
from eureqa.utils.jsonrest import _JsonREST

class SearchLink(_Component):
    """
    Create a link to a specified search

    Example of using HtmlBlock with a SearchLink::

        vl = SearchLink(search)
        h = HtmlBlock('Formatted <b>Card</b> in the Analysis, with reference to search {0}'.format(analysis.html_ref(vl)))
        analysis.create_card(h)


    :param Search search: the search to link to
    :param str link_text: (optional) the text to use for the search link is displayed
    """

    _component_type_str = 'SEARCH_LINK'

    def __init__(self, search=None, link_text=None, _analysis=None,  _component_id=None, _component_type=None):

        # implementation details:
        # self.datasource and self.search are properties
        # self._datasource and self._search is the actual datasource/search object  which may or may not be fetched
        # self._datasource_id and self._search_id are strings and

        if search is not None:
            self.search = search #invoke setter to set all fields

        if link_text is not None:
            self._link_text = link_text
        elif search is not None:
            self._link_text = search.name


        super(SearchLink, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)

    @property
    def search(self):
        """The search linked to

        :rtype: Search
        """
        # Note that the _search field is set by the underlying _Component
        # based on the _datasource_id string
        if hasattr(self, "_eureqa"):
            return self._eureqa._get_search_by_id(self._datasource_id, self._search_id)

    @search.setter
    def search(self, search):
        self._datasource_id = search._data_source_id
        self._search_id = search._id
        self._eureqa = search._eureqa
        self._update()

    @property
    def link_text(self):
        """The text to use for the search link when displayed.

        :return: the link text of this card
        :rtype: str"""
        return self._link_text

    @link_text.setter
    def link_text(self, val):
        self._link_text = val
        self._update()


    def _fields(self):
        return super(SearchLink, self)._fields() + [ 'datasource_id', 'search_id', 'link_text' ]



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import uuid
from eureqa.utils.jsonrest import _JsonREST

# Utility function used in this file
def _iterator_not_empty(it):
    # Python idiom:
    # If you can retrieve at least one element from an iterator,
    # then that iterator is not empty.
    # For many iterators/generators, `len(it) == 0` requires
    # consuming the whole iterator in order to count its elements.
    # So, this is O(1) where that is O(n).
    try:
        it.next()
        return True
    except StopIteration:
        return False

# Python Metaclass that automatically tracks all known types of Components
# If you're in this file, you probably want to skip ahead to 'class Component()' below.
# The following code has a very specific use case.
class _ComponentRegistrar(type):
    """ For internal use only. """

    # For those not familiar with Python metaclasses and __new__:
    #
    # type.__new__() runs every time when a *class* object is created.
    # One can intercept that call by inheriting from type and overriding type.__new__().
    # And then by "registering" metaclass using __metaclass__ attribute of the actual class.
    # Just like we do in _Component class below.
    #
    # In the common case (unless you're creating classes on the fly), a class
    # is "constructed" in this sense when Python first opens and parses the
    # file containing this class.
    #
    # The "Component" class represents a common base class for a
    # collection of serializable objects.
    # The serialized representation of the class contains a 'type' string
    # that indicates what type of thing has been serialized; therefore,
    # which Python class should be used to represent it.
    # In order to de-serialize such objects, 'Component' must therefore
    # know about all of its subclasses that represent specific
    # types of objects, as well as which class represents which 'type'
    # string.
    # Rather than write and maintain a list, this code makes the Python
    # interpreter generate the list for us.

    def __new__(type_, name, bases, dict_):
        cls = type.__new__(type_, name, bases, dict_)
        assert hasattr(cls, "_component_type_str"), \
            "New Analysis Component classes must define a '_component_type' field"
        cls._register_type(cls._component_type_str)
        return cls


class _Component(_JsonREST):
    """
    Component base class.
    Implements common logic shared by all components.
    Do not construct this class directly; instead,
    construct the specific component that you want.

    Notes for component implementors:  (If your new class's sub-Components aren't showing up properly, read on!)

    The backend first sees a Component when the first of the following happens:
      * The Component is constructed with an Analysis passed into the constructor as an argument
      * The Component is added to an Item that is already attached to an Analysis
      * The Component is added to an Item that is not already attached to an Analysis, and then the Item is attached to an Analysis
      * The Component is added to a Table and that Table's data is uploaded
    If more than one of these things happens in the lifecycle of a Component, or if one happens repeatedly,
    only the very first occurrence counts.
    Call this first event "Association".

    Before Association, each individual Component subclass is responsible for tracking their sub-Components (if any).
    At Association, "Component._associate()" is called; each Component that holds sub-Components must override this
    method to also call "_associate()" on all of their sub-Components.
    After Association (and *NOT* before it), each Component is responsible for immediately calling "_associate()"
    on any sub-Component that is added to it.
    """

    __metaclass__ = _ComponentRegistrar

    def clone(self):
        # Make a copy of ourselves
        body = self._to_json()

        # Disassociate the copy from the current Analysis
        if "analysis_id" in body:
            del body["analysis_id"]
        if "component_id" in body:
            del body["component_id"]

        # Instantiate a new Item with this state
        return _Component._construct_from_json(body)

    def delete(self):
        """ Remove this Component from the server. """
        # Subclasses may override this to do more elaborate clean-up.
        self._delete()

    def __init__(self, _analysis=None, _component_id=None, _component_type=None):
        """ For internal use only. """
        if _component_id is not None:
            self._component_id = _component_id

        if _component_type is not None:
            self._component_type = _component_type
        else:
            self._component_type = self._component_type_str

        if _analysis:
            self._associate(_analysis)

    _component_type_str = "GENERIC_COMPONENT"
    _registered_types = {}

    @classmethod
    def _register_type(cls, component_type):
        cls._registered_types[component_type] = cls

    def _populate_eureqa(self):
        if not hasattr(self, "_analysis"):
            # Can't populate Eureqa yet
            return

        self._eureqa = self._analysis._eureqa


    # Override "_construct_from_json()" to use the appropriate subclass if possible
    @classmethod
    def _construct_from_json(cls, body, *args, **kwargs):

        # If we're given an analysis, pass it in after
        # we're done populating the object so that we don't make
        # extraneous HTTP requests populating partial object state.
        analysis = None
        if "_analysis" in kwargs:
            analysis = kwargs["_analysis"]
            del kwargs["_analysis"]

        ret = cls._registered_types.get(body.get("component_type"), _Component)(*args, **kwargs)
        ret._from_json(body)

        if analysis is not None:
            ret._register(analysis)

        ret._populate_eureqa()
        return ret

    # Override "_get_all_from_self()" to associate all newly-constructed components
    # with our Analysis (if any)
    def _get_all_from_self(self):
        """
        Fetch a list of all objects in the same directory as this object.
        :return: list() of objects of the current type
        """
        comps = super(_Component, self)._get_all_from_self()

        if hasattr(self, "_analysis"):
            for comp in comps:
                for c in comp._walk():
                    c._register(self._analysis)

        return comps

    def _object_endpoint(self):
        return '/analysis/%s/components/%s' % (self._analysis._id, self._component_id)

    def _directory_endpoint(self):
        return '/analysis/%s/components' % (self._analysis._id)

    def _fields(self):
        return [ "analysis_id", "component_id", "component_type" ]

    def _batch_endpoint(self, continue_on_error):
        return self._directory_endpoint() + ("?continue_on_error=true" if continue_on_error else "")

    def _batch_update(self, eureqa, components, continue_on_error):
        json_arr = [x._to_json() for x in components]
        if len(json_arr) == 0:
            return  # Nothing to do here

        new_bodies = eureqa._session.execute(self._batch_endpoint(continue_on_error), 'POST', json_arr)
        for component, body in zip(components, new_bodies):
            if body.get("status") != "error":
                component._from_json(body)

    def _pre_associate(self, analysis):
        """
        Per-Component logic that must be performed immediately before
        associate()'ing the current Component with the server.
        Subclasses may override this method, though most shouldn't need to.
        Typically includes uploading dependent objects that are not Components.
        """
        pass

    def _associate(self, analysis):
        """
        Component-walking and -association, v2.0.

        You should NEVER override '_associate()'!  This is a major change from v1.0.

        Here are the methods that you should override:

         * _walk_children(): Should return a list or (preferably) O(1)-memory generator
           of all of the current component's immediate children
         * _register(analysis): Called after all of this component's children have been
           posted to the server and have _component_ids. Many Layout components in
           particular need to update themselves after their children have been given IDs;
           this is their opportunity to do so.

        There's also a third method that's overridden rarely, but is important when it is needed:
         * _pre_associate(analysis): This method is called on a Component immediately
           before that Component is sent to the server in order to be batch-created;
           after we know what Analysis we are going to upload the Component to.  In
           particular, our File and Image components need to upload their file and image
           data to the server prior to being POST'ed as a Component, so that their Component
           JSON structure can refer to the ID of their File.
        """

        # Pre-associate ourselves so that 'self._directory_endpoint()' can work
        # (Don't call 'self._associate()'; the whole point of this method is to
        # avoid an RPC at each level of its recursion, and we haven't done that yet.)
        self._analysis = analysis

        # Upload any files/images/etc that must be uploaded before we can start associating
        for c in self._walk():
            c._pre_associate(analysis)

        # Register all components with the server
        components = [c for c in self._walk() if not getattr(c, "_component_id", False)]
        self._batch_update(analysis._eureqa, components, True)

        # Walk all components; link them to the analysis in Python
        for c in self._walk():
            c._register(analysis)

        # Every component whose child just got an ID has now been modified with that child's new ID.
        # Force a batch-update to account for this.
        # In case we have nested layouts, repeat as needed so that each layer in the chain gets an ID.
        modified_children = components

        # Iterate at most #components times.
        # We should typically only iterate once or twice.
        # We should never actually iterate #components times;
        # the most pedantic edge case should be #components - 1 loops.
        # But, if there's a bug somewhere in this loop logic,
        # avoid an infinite loop.
        for i in xrange(len(components)):
            updated_component_ids = {c._component_id for c in modified_children
                                                      if hasattr(c, "_component_id")}

            modified_parents = \
                [
                    c for c in self._walk()
                      if _iterator_not_empty(
                          child for child in c._walk_children()
                                if getattr(child, "_component_id", None) in updated_component_ids)
                ]

            if len(modified_parents) == 0:
                break

            for c in modified_parents:
                c._register(analysis)

            self._batch_update(analysis._eureqa, modified_parents, True)

            modified_children = modified_parents

        # If, after that, any of our children don't have IDs, that's probably bad data;
        # we should probably ask the server again and error out if necessary.
        components = [c for c in self._walk() if not getattr(c, "_component_id", False)]
        if len(components) != 0:
            self._batch_update(analysis._eureqa, components, False)

    def _to_json(self):
        resp = super(_Component, self)._to_json()

        if hasattr(self, "_analysis"):
            resp["analysis_id"] = self._analysis._id

        return resp

    def _from_json(self, body):
        if hasattr(self, "_analysis") and body.get("analysis_id"):
            assert body.get("analysis_id") == self._analysis._id, \
                "_from_json() can't de-serialize an Item that belongs to a different Analysis"

        super(_Component, self)._from_json(body)

    def _walk(self):
        """
        _walk() returns a generator that iterates over the full Component tree rooted
        at the current node. The order of iteration is not guaranteed other than that
        parents come before their children.

        If you're working with component serialization in the future, you shouldn't
        override this method, but you may use it a lot.
        You *should* override '_walk_children()', below.
        """
        yield self
        for c in self._walk_children():
            for descendent in c._walk():
                yield descendent

    def _walk_children(self):
        return iter([])

    def _register(self, analysis):
        self._analysis = analysis
        self._analysis_id = analysis._id
        _JsonREST.__init__(self, analysis._eureqa)
        self._populate_eureqa()

    def _associate_with_table(self, analysis):
        # The base class implementation of this method only generates UUID but
        # does not add them to the actual analysis. Because most components
        # are directly inlined into the table and some involve extra actions (e.g. uploading a file)
        # It is up to a particular component to override this method and provide more
        # specific implementation.
        if not hasattr(self, "_component_id"):
            self._component_id = str(uuid.uuid4())
        return [self._to_json()]

class _TwoVariablePlotComponent(_Component):
    _component_type_str = "TwoVariablePlot helper mixin (not a complete type)"


    def __init__(self, _analysis=None, _component_id=None, _component_type=None):

        super(_TwoVariablePlotComponent, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)

        default_x_label = getattr(self, '_x_var', 'the x var')
        default_y_label = getattr(self, '_y_var', 'the y var')

        DEFAULTS = {'BOX_PLOT':             {'default_label' : { 'x': default_x_label, 'y': default_y_label },
                                             'default_format': { 'x': 'g',   'y': '.2s' }},

                    'DOUBLE_HISTOGRAM_PLOT':{'default_label' : { 'x': [default_x_label, default_y_label], 'y': 'number of values' },
                                             'default_format': { 'x': 'g', 'y': '.3s' }},

                    'SCATTER_PLOT':         {'default_label' : { 'x': default_x_label, 'y': default_y_label },
                                             'default_format': { 'x': '.3s', 'y': '.3s' }},

                    'BINNED_MEAN_PLOT':     {'default_label' : { 'x': default_x_label, 'y': default_y_label },
                                             'default_format': { 'x': '.3s', 'y': '.2f' }}}



        default_values = DEFAULTS.get(self._component_type,
                                      {'default_label' : { 'x': default_x_label, 'y': default_y_label },
                                       'default_format': { 'x': '.3s', 'y': '.3s' }})

        if not hasattr(self,'_axisLabels'):
            self._axisLabels = default_values['default_label']
        if not hasattr(self,'_labelFormat'):
            self._labelFormat = default_values['default_format']


    class XYMap(object):
        """ A named tuple with keys 'x' and 'y'

        :param TwoVariablePlotCard tvp: The TwoVariablePlotCard object.
        :param dict dic: X and Y parameters as dictionary.
        """

        def __init__(self, parent_component, body):
            self._parent_component = parent_component
            self._body = body

        @property
        def x(self):
            """ X value """
            return self._body['x']

        @x.setter
        def x(self, val):
            self._body['x'] = val
            self._update()

        @property
        def y(self):
            """ Y value """
            return self._body['y']

        @y.setter
        def y(self, val):
            self._body['y'] = val
            self._update()

        def _update(self):
            self._parent_component._update()



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from eureqa.analysis.components.base import _Component
from eureqa.utils.jsonrest import _JsonREST

class HtmlBlock(_Component):
    """Contains free-form user-specified HTML, including references to
    other components.

    Example::

        h = HtmlBlock('This will render on a Card in an Analysis')
        analysis.create_card(h)


    Example of using :class:`HtmlBlock` with a :class:`~eureqa.analysis.components.variable_link.VariableLink`::

        vl = VariableLink(d, 'X')
        h = HtmlBlock('Formatted <b>Card</b> in the Analysis, with reference to {0}'.format(analysis.html_ref(vl)))
        analysis.create_card(h)


    :param str html: Body of the card.

    :var str eureqa.analysis.components.html_block.HtmlBlock.html: Body of the card.

    If you need particular styles or JS libraries, you MUST define
    and/or load them as part of the HtmlBlock where they are needed.
    Any styles or JS objects defined by Eureqa may change or be
    removed from release to release.

    Embedding custom JavaScript (or CSS, which can contain JavaScript)
    from a malicious third-party source can grant that source access
    to your site.  Only embed content from trusted sources.

    """

    _component_type_str = 'ANALYSIS_HTML_BLOCK'

    def __init__(self, html='', _analysis=None, _component_id=None, _component_type=None):
        if html is not None:
            self._html = html

        super(HtmlBlock, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)


    @property
    def html(self):
        """ The body of this card.

        :return: body of this card
        :rtype: str"""
        return self._html

    @html.setter
    def html(self, val):
        self._html = val
        self._update()

    def _fields(self):
        return super(HtmlBlock, self)._fields() + [ 'html' ]



#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
